 /* ===========================================================================
 *   Copyright (C) 2013 by ]init[ AG für digitale Kommunikation
 * -----------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  See also the full LGPL text here: http://www.gnu.org/copyleft/lesser.html
 * ===========================================================================*/
package de.init.loext.its.dialogs.ui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.LayoutManager;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.KeyStroke;
import javax.swing.ListSelectionModel;
import javax.swing.LookAndFeel;
import javax.swing.UIManager;
import javax.swing.UIManager.LookAndFeelInfo;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import com.sun.star.table.BorderLine;

import de.init.its.ds.ItsAttributes;
import de.init.its.ds.ItsDocument;
import de.init.its.ds.ItsAttributes.DataCategories;
import de.init.loext.its.transport.tags.Locales;
import de.init.util.logging.SimpleLog;

/**
 * Collection of all Dialogs
 * 
 * @author mkandora
 *
 */
@SuppressWarnings("all")
public abstract class ItsDialogs {
	
	
	public enum OpStates {
		SAVE, CANCEL, REMOVE;
	}
	
	/**
	 * Small listener delegate for all (save|cancel|remove) operations
	 *  
	 * @author mkandora
	 *
	 */
	public static interface TerminateConfigurationListener
	{
		
		/**
		 * Will be called, whenn save is activated
		 * @param dialog
		 */
		public void doSave(final ItsDialogBase dialog);
		
		/**
		 * Will be called when remove is activated
		 * @param dialog
		 */
		public void doRemove(final ItsDialogBase dialog);
		
		/**
		 * Will be called when cancel is activated
		 * @param dialog
		 */
		public void doCancel(final ItsDialogBase dialog);
		
	}
	
	public static interface StateChangeListener {
		
		/**
		 * called, whenever a state has been changed
		 * 
		 * @param attribute
		 * @param dialog
		 */
		public void doAction(final ItsAttributes attribute, final ItsDialogBase dialog, final String...values);
	}
	
	/**
	 * Base Class for all Dialogs
	 * 
	 * @author mkandora
	 *
	 */
	public static abstract class ItsDialogBase extends JFrame
	{
		protected String title;
		protected ItsDialogModel model;
		protected TerminateConfigurationListener listener;
		protected StateChangeListener stateChangeListener;
		protected JButton save;
		protected JButton cancel;
		protected JButton remove;
		
		static {
			try {
				// set a more nicer looking lookAndFeel for all ui
				for(final LookAndFeelInfo look : UIManager.getInstalledLookAndFeels()) {
					if("Nimbus".equals(look.getName())) {
						UIManager.setLookAndFeel(look.getClassName());
						break;
					}
				}
			} catch(Exception e) {
				e.printStackTrace();
			}
		}
		
		public ItsDialogBase(final String title, final TerminateConfigurationListener sclistener) {
			setTitle(title);
			setDefaultCloseOperation(DISPOSE_ON_CLOSE);
			setResizable(false);
			
			this.model 			= ItsDialogModel.getModel();
			this.title 			= title;
			this.listener 		= sclistener;
			this.save 			= new JButton(OpStates.SAVE.name().toLowerCase());
			this.cancel	 		= new JButton(OpStates.CANCEL.name().toLowerCase());
			this.remove			= new JButton(OpStates.REMOVE.name().toLowerCase());
			
			// provide a basic implementation of this interface. delegate all changes 
			// to this routine
			// 
			// respects previously set values, updates them or creates a new set of values
			this.stateChangeListener = new StateChangeListener() {
				
				@Override
				public void doAction(final ItsAttributes attribute, final ItsDialogBase dialog, final String...values) {
					Map<ItsAttributes, List<String>> mapping = model.getValues(dialog) != null ? 
							model.getValues(dialog) : new HashMap<ItsAttributes, List<String>>();
					mapping.put(attribute, Arrays.asList(values));
					model.setValues(dialog, mapping);
				}
			};
			final JFrame frame = this;
			getRootPane().getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE,0), "ItsAction");
			getRootPane().getActionMap().put("ItsAction", new AbstractAction() {
				@Override
				public void actionPerformed(ActionEvent e) {
					 frame.dispatchEvent(new WindowEvent(frame, WindowEvent.WINDOW_CLOSING )); 
				}
			});
			
			addKeyListener(new KeyAdapter() {
				@Override
				public void keyTyped(KeyEvent e) {
					if(e.getKeyCode() == KeyEvent.VK_ESCAPE) {
						dispose();
					}
				}
			});
			
			SimpleLog.log("Registering EventListeners");
			
			this.save.addActionListener(new ActionListener() {
				
				@Override
				public void actionPerformed(ActionEvent e) {
					listener.doSave(ItsDialogBase.this); 
				}
			});
			this.cancel.addActionListener(new ActionListener() {
				
				@Override
				public void actionPerformed(ActionEvent e) {
					listener.doCancel(ItsDialogBase.this); 
				}
			});
			this.remove.addActionListener(new ActionListener() {
				
				@Override
				public void actionPerformed(ActionEvent e) {
					listener.doRemove(ItsDialogBase.this); 
				}
			});
			
			this.addWindowListener(new WindowAdapter() {
				@Override
				public void windowClosing(WindowEvent e) {
					super.windowClosing(e);
					listener.doCancel(ItsDialogBase.this);
					reset();
				}
			});
		}
		
		protected JPanel getOpPanel() {
			JPanel panel 	= new JPanel();
			
			panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));
			panel.setBorder(BorderFactory.createEtchedBorder());
			panel.add(save);
			panel.add(remove);
			panel.add(cancel);
			save.setAlignmentX(Component.RIGHT_ALIGNMENT);
			cancel.setAlignmentX(Component.RIGHT_ALIGNMENT);
			return panel;
		}
		
		protected JPanel createTextField() {
			return null;
		}
		
		protected JPanel createRadioButtons() {
			return null;
		}
		
		protected ItsDialogBase addToContentPane(final JComponent comp, Object hints) {
			getContentPane().add(comp, hints);
			return this;
		}
		
		/**
		 * displays this component
		 */
		public abstract void display();
		
		/**
		 * releases this component
		 */
		public void release() {}
		
		public ItsDialogBase update() {
			return this;
		}
		
		public abstract void reset();
		
		/**
		 * Displays a warning dialog message
		 * 
		 * @param message the message to be shown
		 */
		public void displayErrorMessage(final String message) {
			JOptionPane.showMessageDialog(null, message);
		}
	}
	
	@ItsDialogView(category = DataCategories.LOCALEFILTER)
	public static class ItsLocFilterDialog extends ItsDialogBase
	{
		private JList list;
		private JRadioButton include;
		private JRadioButton exclude;
		private JTextField textfield;
		private ButtonGroup bg;

		public ItsLocFilterDialog(String title, TerminateConfigurationListener sclistener) {
			super(title, sclistener);
			
			bg = new ButtonGroup();
			include = new JRadioButton("include");
			exclude = new JRadioButton("exclude");
			include.setAlignmentX(Component.LEFT_ALIGNMENT);
			exclude.setAlignmentX(Component.LEFT_ALIGNMENT);
			bg.add(include);
			bg.add(exclude);
			
			// preselect include
			include.setSelected(true);
			
			JPanel listpanel = new JPanel();
			listpanel.setLayout(new BoxLayout(listpanel, BoxLayout.X_AXIS));
			listpanel.setBorder(BorderFactory.createTitledBorder("Locale Filter List"));
			
			JPanel textPanel = new JPanel();
			textPanel.setLayout(new BoxLayout(textPanel, BoxLayout.Y_AXIS));
			textPanel.setBorder(BorderFactory.createTitledBorder("Locale Filter List"));
			textfield = new JTextField(23);
			textPanel.add(textfield);
			
			// set to true
			textfield.setEditable(true);
		
			
			list = new JList();
			list.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		    list.setPrototypeCellValue("de-de"); //get extra space
		    list.setLayoutOrientation(JList.VERTICAL);
		    list.setVisibleRowCount(-1);
			list.setListData(Locales.getNormalizedLocales());
			
			
			
			JScrollPane scrollPane = new JScrollPane(list);
			scrollPane.setPreferredSize(new Dimension(512, 120));
			listpanel.add(scrollPane);
			
			
			JPanel radiopanel = new JPanel();
			radiopanel.setLayout(new BoxLayout(radiopanel, BoxLayout.Y_AXIS));
			radiopanel.setBorder(BorderFactory.createTitledBorder("Locale Filter Type"));
//			radiopanel.setAlignmentX(Component.LEFT_ALIGNMENT);
			radiopanel.add(include);
			radiopanel.add(exclude);
			
			radiopanel.setMaximumSize(new Dimension(Integer.MAX_VALUE,
                    Integer.MAX_VALUE));
			
			
			JPanel centerPanel = new JPanel();
			centerPanel.setLayout(new BoxLayout(centerPanel, BoxLayout.Y_AXIS));
			centerPanel.add(textPanel);
			centerPanel.add(listpanel);
			
			
			this.
			addToContentPane(radiopanel, BorderLayout.NORTH).
			addToContentPane(centerPanel, BorderLayout.CENTER).
			addToContentPane(getOpPanel(), BorderLayout.SOUTH);
			
			
			
			
			// add action listener;
			include.addActionListener(new ActionListener() {
				
				@Override
				public void actionPerformed(ActionEvent e) {
					stateChangeListener.doAction(ItsAttributes.localeFilterType, ItsLocFilterDialog.this, "include");
				}
			});
			
			exclude.addActionListener(new ActionListener() {
				
				@Override
				public void actionPerformed(ActionEvent e) {
					stateChangeListener.doAction(ItsAttributes.localeFilterType, ItsLocFilterDialog.this, "exclude");
				}
			});
			
			textfield.addActionListener(new ActionListener() {
				
				@Override
				public void actionPerformed(ActionEvent e) {
					stateChangeListener.doAction(ItsAttributes.localeFilterList, ItsLocFilterDialog.this, textfield.getText());
				}
			});
			
			textfield.addKeyListener(new KeyAdapter() {
				@Override
				public void keyReleased(KeyEvent e) {
					// TODO Auto-generated method stub
					stateChangeListener.doAction(ItsAttributes.localeFilterList, ItsLocFilterDialog.this, textfield.getText().trim());
				}
			});
			
			list.addListSelectionListener(new ListSelectionListener() {
				
				@Override
				public void valueChanged(ListSelectionEvent e) {
					StringBuffer buffer = new StringBuffer();
					int index = 0;
					for(int i : list.getSelectedIndices()) {
						if(index > 0)
						buffer.append(", ");
						buffer.append(Locales.getNormalizedLocale(i));
						++index;
					}
					textfield.setText(buffer.toString());
					stateChangeListener.doAction(ItsAttributes.localeFilterList, ItsLocFilterDialog.this, buffer.toString().trim());
				}
			});
		}

		@Override
		public void display() {
			
			setLocationRelativeTo(null);
			pack();
			setVisible(true);
		}

		@Override
		public ItsDialogBase update() {
			for(ItsAttributes its : model.getValues(this).keySet()) {
				switch(its) {
				case localeFilterList:
					textfield.setText(model.getValues(this).get(its).toString().replaceAll("(\\[|\\])", ""));
					break;
				case localeFilterType:
					exclude.setSelected(model.getValues(this).get(its).contains("exclude"));
					include.setSelected(model.getValues(this).get(its).contains("include"));
					break;
				}
			}
			return this;
		}

		@Override
		public void reset() {
			textfield.setText("");
			list.clearSelection();
//			include.setSelected(false);
//			exclude.setSelected(false);
			bg.clearSelection();
		}

	}
	
	@ItsDialogView(category = DataCategories.LOCALIZATIONNOTE)
	public static class ItsLocNoteDialog extends ItsDialogBase
	{

		private JRadioButton alert;
		private JRadioButton description;
		private JPanel textPanel;
		private JTextArea locNote;
		private JTextField textfield;
		private ButtonGroup bg;

		public ItsLocNoteDialog(String title, TerminateConfigurationListener sclistener) {
			super(title, sclistener);
			
			bg = new ButtonGroup();
			alert = new JRadioButton("alert");
			description = new JRadioButton("description");
			alert.setAlignmentX(Component.LEFT_ALIGNMENT);
			description.setAlignmentX(Component.LEFT_ALIGNMENT);
			bg.add(alert);
			bg.add(description);
			description.setSelected(true);
			JPanel radiopanel = new JPanel();
			radiopanel.setLayout(new BoxLayout(radiopanel, BoxLayout.Y_AXIS));
			radiopanel.setBorder(BorderFactory.createTitledBorder("Localization Note Type"));
			radiopanel.add(alert);
			radiopanel.add(description);
			
			radiopanel.setMaximumSize(new Dimension(Integer.MAX_VALUE,
                    Integer.MAX_VALUE));
			
			
			textPanel = new JPanel();
			textPanel.setLayout(new BoxLayout(textPanel, BoxLayout.Y_AXIS));
			textPanel.setBorder(BorderFactory.createTitledBorder("Localization Note URI"));
			textfield = new JTextField(23);
			textPanel.add(textfield);
			textfield.setEditable(true);
			
			/// -----------------------------------------------------------------
			
			JPanel listpanel = new JPanel();
			listpanel.setLayout(new BoxLayout(listpanel, BoxLayout.X_AXIS));
			listpanel.setBorder(BorderFactory.createTitledBorder("Localization Note"));
			locNote = new JTextArea(5, 23);
			JScrollPane scrollPane = new JScrollPane(locNote);
			scrollPane.setPreferredSize(new Dimension(512, 120));
			listpanel.add(scrollPane);
			
			/// -----------------------------------------------------------------
			
			JPanel centerPanel = new JPanel();
			centerPanel.setLayout(new BoxLayout(centerPanel, BoxLayout.Y_AXIS));
			centerPanel.add(textPanel);
			centerPanel.add(listpanel);
			
			addToContentPane(radiopanel, BorderLayout.NORTH).
			addToContentPane(centerPanel, BorderLayout.CENTER).
			addToContentPane(getOpPanel(), BorderLayout.SOUTH);
			
			// listeners
			alert.addActionListener(new ActionListener() {
				
				@Override
				public void actionPerformed(ActionEvent e) {
					stateChangeListener.doAction(ItsAttributes.locNoteType, ItsLocNoteDialog.this, "alert");
				}
			});
			
			description.addActionListener(new ActionListener() {
				
				@Override
				public void actionPerformed(ActionEvent e) {
					stateChangeListener.doAction(ItsAttributes.locNoteType, ItsLocNoteDialog.this, "description");
				}
			});
			
			textfield.addKeyListener(new KeyAdapter() {
				
				@Override
				public void keyReleased(KeyEvent e) {
					if(!locNote.getText().isEmpty()) {
						displayErrorMessage("You can only set a localization Note Reference or a Localization Node");
						textfield.setText("");
						return;
					}
					stateChangeListener.doAction(ItsAttributes.locNoteRef, ItsLocNoteDialog.this, textfield.getText()); 
				}
			});
			
			locNote.addKeyListener(new KeyAdapter() {
				@Override
				public void keyReleased(KeyEvent e) {
					if(!textfield.getText().isEmpty()) {
						displayErrorMessage("You can only set a localization Note Reference or a Localization Node");
						locNote.setText("");
						return;
					}
					stateChangeListener.doAction(ItsAttributes.locNote, ItsLocNoteDialog.this, locNote.getText() ); 
				}
			});
		}

		@Override
		public void display() {
			setLocationRelativeTo(null);
			pack();
			setVisible(true);
		}
		
		@Override
		public ItsDialogBase update() {
			for(ItsAttributes its : model.getValues(this).keySet()) {
				switch(its) {
				case locNote:
					locNote.setText(model.getValues(this).get(its).toString().replaceAll("(\\[|\\])", ""));
					break;
				case locNoteRef:
					textfield.setText(model.getValues(this).get(its).toString().replaceAll("(\\[|\\])", ""));
					break;
				case locNoteType:
					alert.setSelected(model.getValues(this).get(its).contains("alert"));
					description.setSelected(model.getValues(this).get(its).contains("description"));
					break;
				}
			}
			return this;
		}
		
		@Override
		public void reset() {
			bg.clearSelection();
			textfield.setText("");
			locNote.setText("");
		}
		
	}
	
	@ItsDialogView(category = DataCategories.TERMINOLOGY)
	public static class ItsTerminologyDialog extends ItsDialogBase
	{


		private JRadioButton yes;
		private JRadioButton no;
		private JTextField termInfoRef;
		private JTextField termConfidence;
		private ButtonGroup bg;

		public ItsTerminologyDialog(String title, TerminateConfigurationListener sclistener) {
			super(title, sclistener);
			// TODO Auto-generated constructor stub
			
			bg = new ButtonGroup();
			yes = new JRadioButton("yes");
			no = new JRadioButton("no");
			yes.setAlignmentX(Component.LEFT_ALIGNMENT);
			no.setAlignmentX(Component.LEFT_ALIGNMENT);
			bg.add(yes);
			bg.add(no);
			no.setSelected(true);
			JPanel radiopanel = new JPanel();
			radiopanel.setLayout(new BoxLayout(radiopanel, BoxLayout.Y_AXIS));
			radiopanel.setBorder(BorderFactory.createTitledBorder("Is term?"));
			radiopanel.add(yes);
			radiopanel.add(no);
			
			radiopanel.setMaximumSize(new Dimension(Integer.MAX_VALUE,
                    Integer.MAX_VALUE));
			// -----------------------------------------------------------------
			
			JPanel textPanel = new JPanel();
			textPanel.setLayout(new BoxLayout(textPanel, BoxLayout.Y_AXIS));
			textPanel.setBorder(BorderFactory.createTitledBorder("Term Information Reference"));
			termInfoRef = new JTextField(23);
			textPanel.add(termInfoRef);
			termInfoRef.setEditable(true);
			// -----------------------------------------------------------------
			
			JPanel termConfidencePanel = new JPanel();
			termConfidencePanel.setLayout(new BoxLayout(termConfidencePanel, BoxLayout.Y_AXIS));
			termConfidencePanel.setBorder(BorderFactory.createTitledBorder("Term Confidence [0.0-1.0]"));
			termConfidence = new JTextField(23);
			termConfidencePanel.add(termConfidence);
			termInfoRef.setEditable(true);
			// -----------------------------------------------------------------
			
			JPanel centerPanel = new JPanel();
			centerPanel.setLayout(new BoxLayout(centerPanel, BoxLayout.Y_AXIS));
			centerPanel.add(textPanel);
			centerPanel.add(termConfidencePanel);
			
			addToContentPane(radiopanel, BorderLayout.NORTH).
			addToContentPane(centerPanel, BorderLayout.CENTER).
			addToContentPane(getOpPanel(), BorderLayout.SOUTH);
			
			// listeners
			yes.addActionListener(new ActionListener() {
				
				@Override
				public void actionPerformed(ActionEvent e) {
					stateChangeListener.doAction(ItsAttributes.term, ItsTerminologyDialog.this, "yes");
				}
				
			});
			
			no.addActionListener(new ActionListener() {
				
				@Override
				public void actionPerformed(ActionEvent e) {
					stateChangeListener.doAction(ItsAttributes.term, ItsTerminologyDialog.this, "no");
				}
			});
			
			termInfoRef.addKeyListener(new KeyAdapter() {
				@Override
				public void keyReleased(KeyEvent e) {
					stateChangeListener.doAction(ItsAttributes.termInfoRef, ItsTerminologyDialog.this, termInfoRef.getText());
				}
			});
			
			termConfidence.addKeyListener(new KeyAdapter() {
				@Override
				public void keyReleased(KeyEvent e) {
					String text = termConfidence.getText();
					if(text.matches("([0-1]?\\.?\\d{0,8})?")) { 
						stateChangeListener.doAction(ItsAttributes.termConfidence, ItsTerminologyDialog.this, termConfidence.getText());
					} else {
						termConfidence.setText("");
					}
				}
			});
		}

		@Override
		public void display() {
			setLocationRelativeTo(null);
			pack();
			setVisible(true);
		}
		
		@Override
		public ItsDialogBase update() {
			if(model.getValues(this) == null) return this;
			for(ItsAttributes its : model.getValues(this).keySet()) {
				switch(its) {
				case term:
					yes.setSelected(model.getValues(this).get(its).contains("yes"));
					no.setSelected(model.getValues(this).get(its).contains("no"));
					break;
				case termConfidence:
					termConfidence.setText(model.getValues(this).get(its).toString().replaceAll("(\\[|\\])", ""));
					break;
				case termInfoRef:
					termInfoRef.setText(model.getValues(this).get(its).toString().replaceAll("(\\[|\\])", ""));
					break;
				}
			}
			return this;
		}

		@Override
		public void reset() {
			bg.clearSelection();
			termConfidence.setText("");
			termInfoRef.setText("");
		}

		
	}
	
	@ItsDialogView(category = DataCategories.TRANSLATE)
	public static class ItsTranslateDialog extends ItsDialogBase
	{

		private JRadioButton yes;
		private JRadioButton no;
		private ButtonGroup bg;

		public ItsTranslateDialog(String title, TerminateConfigurationListener sclistener) {
			super(title, sclistener);
			
			bg = new ButtonGroup();
			yes = new JRadioButton("yes");
			no = new JRadioButton("no");
			yes.setAlignmentX(Component.LEFT_ALIGNMENT);
			no.setAlignmentX(Component.LEFT_ALIGNMENT);
			bg.add(yes);
			bg.add(no);
			no.setSelected(true);
			
//			// push attributes set
//			model.getValues(this).put(ItsAttributes.translate, Arrays.asList("no"));
			
			JPanel radiopanel = new JPanel();
			radiopanel.setLayout(new BoxLayout(radiopanel, BoxLayout.Y_AXIS));
			radiopanel.setBorder(BorderFactory.createTitledBorder("Translate?"));
			radiopanel.add(yes);
			radiopanel.add(no);
			
			radiopanel.setMaximumSize(new Dimension(Integer.MAX_VALUE,
                    Integer.MAX_VALUE));
                    
            addToContentPane(radiopanel, BorderLayout.NORTH).
            addToContentPane(getOpPanel(), BorderLayout.SOUTH);
            
            /// listeners
            yes.addActionListener(new ActionListener() {
				
				@Override
				public void actionPerformed(ActionEvent e) {
					stateChangeListener.doAction(ItsAttributes.translate, ItsTranslateDialog.this, "yes");
				}
			});
            
            no.addActionListener(new ActionListener() {
				
				@Override
				public void actionPerformed(ActionEvent e) {
					stateChangeListener.doAction(ItsAttributes.translate, ItsTranslateDialog.this, "no");
				}
			});
		}

		@Override
		public void display() {
			
			if(model.getValues(this).isEmpty()) {
				model.getValues(this).put(ItsAttributes.translate, Arrays.asList("no"));
			}
			
			setLocationRelativeTo(null);
			pack();
			setVisible(true);
		}
		
		@Override
		public ItsDialogBase update() {
			for(ItsAttributes its : model.getValues(this).keySet()) {
				switch(its) {
				case translate:
					yes.setSelected(model.getValues(this).get(its).get(0).equals("yes"));
					no.setSelected(model.getValues(this).get(its).get(0).equals("no"));
					break;
				}
			}
			return this;
		}

		@Override
		public void reset() {
			// TODO Auto-generated method stub
			bg.clearSelection();
			
			// set 'no' as default selected
			no.setSelected(true);
		}
		
	}

}
