 /* ===========================================================================
 *   Copyright (C) 2013 by ]init[ AG für digitale Kommunikation
 * -----------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  See also the full LGPL text here: http://www.gnu.org/copyleft/lesser.html
 * ===========================================================================*/
package de.init.service;

import com.sun.star.lang.XMultiServiceFactory;
import com.sun.star.uno.RuntimeException;
import com.sun.star.uno.UnoRuntime;
import com.sun.star.uno.XComponentContext;
import com.sun.star.uno.XInterface;

/**
 * Wrapper class for UNO services and interfaces
 * 
 * @author mkandora
 * 
 */
public class UnoServiceManager {

	private static UnoServiceManager manager;
	private XComponentContext context;
	private XMultiServiceFactory servicefactory;

	private UnoServiceManager(final XComponentContext context) {
		this.context = context;
		this.servicefactory = query(XMultiServiceFactory.class, this.context.getServiceManager());
	}

	/**
	 * Obtains a singleton instance of a UnoServiceManager using the context
	 * 
	 * @param context
	 * @return
	 */
	public static UnoServiceManager getServiceManager(
			final XComponentContext... context) {
		if (context == null && manager == null)
			throw new RuntimeException(
					"Manager has not been initialized! Please provide a valid non-null context.");
		return manager != null ? manager : (manager = new UnoServiceManager(
				context[0]));
	}
	
	/**
	 * Queries the {@link UnoRuntime} for provided Interface on standard
	 * {@link XComponentContext} 
	 * 
	 * @param clazz the {@link XInterface} to query the standard {@link XComponentContext} for
	 * @return the {@link XInterface}
	 */
	public final <T extends XInterface> T queryWithDefaultContext(final Class<T> clazz) {
		return UnoRuntime.queryInterface(clazz, context);
	}
	
	/**
	 * Queries the {@link UnoRuntime} for provided Interface on provided context
	 * 
	 * @param clazz the {@link XInterface} to query the standard {@link XComponentContext} for
	 * @param context a context to query the runtime for interfaces
	 * @return the {@link XInterface}
	 */
	public final <T extends XInterface> T query(final Class<T> clazz, final Object context) {
		return UnoRuntime.queryInterface(clazz, context);
	}
	
	
	/**
	 * Queries the {@link XComponentContext} for a given service
	 * 
	 * The {@link XMultiServiceFactory} from {@link XComponentContext#getServiceManager()} will be used to
	 * create new services
	 * 
	 * @param serviceName
	 * @return
	 * @throws Exception
	 */
	public final Object getService(final String serviceName) throws Exception {
		return this.servicefactory.createInstance(serviceName);
	}
	
	public final Object getService(final String serviceName, final XMultiServiceFactory serviceFactory) throws Exception {
		return serviceFactory.createInstance(serviceName);
	}

}
