 /* ===========================================================================
 *   Copyright (C) 2013 by ]init[ AG für digitale Kommunikation
 * -----------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  See also the full LGPL text here: http://www.gnu.org/copyleft/lesser.html
 * ===========================================================================*/

package de.init.its.api.process;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.StringReader;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import de.init.its.ds.ItsAttributes;
import de.init.its.ds.ItsAttributes.DataCategories;
import de.init.its.util.Util;

/**
 * ODT File styling configuration. This class may be used to 
 * load the correct styling for all implemented ITS Data Categories 
 * inside a LibreOffice Writer Document. 
 * 
 * This class provides the styling configuration in any case.
 * It is assumed that a CSS like file name <b>mlw-global.css</b> exists and
 * tries to load it, if this file does not exist or has incorrect syntax
 * a default configuration is provided:
 * 
 * <pre>
 *.translate 
 *{                                                                                   
 *	background-color: #aaffaa;                                                                
 *}                                                                                         
 *			                                                                                      
 *.localizationnote 
 *{                                                                           
 *	text-overline-style: solid; text-overline-width: bold; text-overline-color: font-color;   
 *}                                                                                             
 *	                                                                                              
 *.terminology 
 *{                                                                                
 *	text-overline-style: solid; text-overline-width: bold; text-overline-color: #ffffff;      
 *}                                                                                             
 *	                                                                                              
 * .localefilter 
 * {                                                                               
 * 	text-overline-style: solid; text-overline-width: bold; text-overline-color: #000099;
 * }      
 */                                                                                           
public class OdtStyleConfiguration
	{
		
		/**
		 * default configuration is as follows
		 * .translate {
				background-color: #aaffaa; 
			}
			
			.localizationnote {
				text-overline-style: solid; text-overline-width: bold; text-overline-color: font-color; 
			}
			
			.terminology {
				text-overline-style: solid; text-overline-width: bold; text-overline-color: #ffffff;
			}
			
			.localefilter {
				text-overline-style: solid; text-overline-width: bold; text-overline-color: #000099;
			}
		 */
		
		/**
		 * reflects the mapping between ITS data categories and the mapping between style name and value
		 * 
		 *  {:DataCategories.TRANSLATE {:stylename stylevalue} :DataCategories:LOCALIZATIONNOTE {:stylename stylevalue}}
		 */
		private static Map<String, Map<String, String>> odtStyleMapping;
		
		/**
		 * Reflects the mapping between ODT File Styles and the UNO Character Styles
		 * {:CharBackColor {background-color}}
		 */
		private static Map<String, String> unoStyleMapping;
		
		private static Map<DataCategories, Map<String, String>> genMapping;
		
		private static final String stylingFile = "/styles/mlw-global.css";
		
		
		
		public static void load(URL resource) {
			try {
				genMapping 		= new HashMap<DataCategories, Map<String, String>>();
				unoStyleMapping = new HashMap<String, String>();
				odtStyleMapping = new HashMap<String, Map<String, String>>();
				loadConfig(resource);
			} catch (Exception e) {
//				// TODO Auto-generated catch block
				loadDefaultConfig();
			}
			
			// manual mapping. no other alternative found
			unoStyleMapping.put("CharBackColor", "background-color");
			unoStyleMapping.put("CharUnderline", "text-underline-style");
			unoStyleMapping.put("CharUnderlineColor", "text-underline-color");
			unoStyleMapping.put("CharOverline", "text-overline-style");
			unoStyleMapping.put("CharOverlineColor","text-overline-color");
			
			for(DataCategories dc : DataCategories.values())
			{
				for(String unoName : unoStyleMapping.keySet()) {
					
					if(odtStyleMapping.get(dc.name()).containsKey(unoStyleMapping.get(unoName)))
					{
						String value = odtStyleMapping.get(dc.name()).get(unoStyleMapping.get(unoName));
						if(!genMapping.containsKey(dc)) {
							Map<String, String> values = new HashMap<String, String>();
							values.put(unoName, value);
							genMapping.put(dc, values);
						}
						else {
							Map<String, String> values = genMapping.get(dc);
							values.put(unoName, value);
							genMapping.put(dc, values);
						}
					}
				}
			}
			
			try {
			} catch(Exception e) {
				e.printStackTrace();
			}
		}
		
		/**
		 * if loading an external files fails, initialise the style mapping using these hard-coded styles
		 */
		protected static synchronized void loadDefaultConfig() {
			// translate
			Map<String, String> translateStyles = new HashMap<String, String>();
			translateStyles.put("background-color", "#aaffaa");
			odtStyleMapping.put(ItsAttributes.DataCategories.TRANSLATE.name(), translateStyles);
			
			// localization note
			Map<String, String> localizationnoteStyles = new HashMap<String, String>();
//			localizationnoteStyles.put("background-color", "#bad666");
			localizationnoteStyles.put("text-underline-width", "bold");
			localizationnoteStyles.put("text-underline-color", "font-color");
			localizationnoteStyles.put("text-underline-style", "solid");
			odtStyleMapping.put(ItsAttributes.DataCategories.LOCALIZATIONNOTE.name(), localizationnoteStyles);
			
			// terminology 
			Map<String, String> terminologyStyles = new HashMap<String, String>();
			terminologyStyles.put("background-color", "#ffaaaa");
//			terminologyStyles.put("text-overline-width", "bold");
//			terminologyStyles.put("text-overline-color", "#ffffff");
//			terminologyStyles.put("text-overline-style", "solid");
			odtStyleMapping.put(ItsAttributes.DataCategories.TERMINOLOGY.name(), terminologyStyles);
			
			// localefilter styles
			Map<String, String> localefilterStyles = new HashMap<String, String>();
//			localefilterStyles.put("background-color", "#00ff00");
			localefilterStyles.put("text-overline-width", "bold");
			localefilterStyles.put("text-overline-color", "#000099");
			localefilterStyles.put("text-overline-style", "solid");
			odtStyleMapping.put(ItsAttributes.DataCategories.LOCALEFILTER.name(),localefilterStyles);
		}
		
		/**
		 * two pass, no lookahead simple css file parser routine. loads an externally 
		 * configured css file named 'mlw-global.css'.
		 * 
		 * @throws Exception
		 */
		protected static synchronized void loadConfig(URL resource) throws Exception {
			// FIXME:  WARNING check if resource is valid and accessible via UNO Context!!!
			URL resource_ = ClassLoader.getSystemClassLoader().getResource(stylingFile);
			
			BufferedReader reader = 
				new BufferedReader(new FileReader(new File(resource.getFile())));
			String line = "";
			int inComment = 0; // higher then 0 is inComment, 0 means data/code
			
			StringBuffer buffer = new StringBuffer();
			
			// pass 1: read everything into string
			while((line = reader.readLine()) != null) {
				if(line.startsWith("/*")) {
					inComment++;
				}
				else if(line.endsWith("*/")) {
					inComment--;
					continue;
				}
				if(inComment < 1) { // continue lexing job, data blocks are available
					buffer.append(line);
				}
			} // end pass 1
			
			// close the reading process
			reader.close();
			
			// define state variables for lexing
			int inBlock = 0;
			int isKey = 0; // 0 is key, 1 is value
			char nextChar = '\0';
			int next = 0;
			String lastDCname 	= "";
			String lastKey 		= "";
			String lastValue 	= "";
			
			Map<String, String> lastKVMapping = null;
			
			StringReader stringReader = new StringReader(buffer.toString());
			// pass 2: parse values from css config
			while((next =  stringReader.read()) != -1) {
				
				// cast to char now, doing this a priori will automatically convert -1 to 65535 (short max) for chars
				nextChar = (char)next;
				if(inBlock > 2 ) {
					// break on nested blocks
					throw new Exception("syntax error. incorrect nesting of values for style def block");
				}
				switch(nextChar)
				{
				// drop whitespace characters
				case ' ':
				case '\t':
				case '\r':
				case '\n':
					continue;
				case '{':
					// add the currently found datacategory style confguration name
					// to the style mapping
					if((lastKVMapping = odtStyleMapping.get(lastDCname)) == null) {
						lastKVMapping = new HashMap<String, String>();
					}
					odtStyleMapping.put(lastDCname.toUpperCase(), lastKVMapping);
					inBlock++;
					if(inBlock > 2) throw new Exception("syntax error. incorrect nesting of values for style def block");
					break;
				case '}':
					lastDCname = "";
					inBlock--;
					break;
				case ';':
					lastKVMapping = odtStyleMapping.get(lastDCname.toUpperCase());
					lastKVMapping.put(lastKey, lastValue);
					lastKey = lastValue = "";
					isKey = 0;
					break;
				case ':':
					isKey = 1;
					break;
				default:
					if(inBlock > 0) {
						if(isKey < 1) {
							// can only be key
							lastKey += nextChar;
						}
						else {
							// can only be the value
							lastValue += nextChar;
						}
					}
					else {
						lastDCname += "" + nextChar;
					}
					break;
				}
			}
			// close resource
			reader.close();
		}
		
		/**
		 * returns the style mapping (wether the externally configured one, or a default mapping, 
		 * if external file contains syntax errors)
		 * 
		 * @return a mapping between the data category name {@linkplain ItsAttributes.DataCategories} and the style values
		 * applied for an odt file reconversion
		 */
		public static Map<String, Map<String, String>> getStyleMapping() {
			return odtStyleMapping;
		}
		
		/**
		 * Returns the style mapping for a specific {@link DataCategories}
		 * 
		 * @param datacategory the {@link DataCategories} to get the styling
		 * @return the mapping for the styles
		 */
		public static Map<String, String> getStyleMappingForDataCategory(final DataCategories datacategory) {
			return odtStyleMapping != null ? odtStyleMapping.get(datacategory.name()) : new HashMap<String, String>();
		}
		
		public static Map<String, String> getUnoStyleMappingForDataCategory(final DataCategories datacategory) {
			return genMapping.get(datacategory);
		}
		
		public static void main(String[] args) {
			System.out.println(OdtStyleConfiguration.getStyleMapping());
			
			
		}
	}