 /* ===========================================================================
 *   Copyright (C) 2013 by ]init[ AG für digitale Kommunikation
 * -----------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  See also the full LGPL text here: http://www.gnu.org/copyleft/lesser.html
 * ===========================================================================*/
package de.init.its.api.spi;

import java.io.Serializable;
import java.util.List;
import java.util.Map;

import de.init.its.api.ds.IitsAttribute;
import de.init.its.api.ds.IitsId;
import de.init.its.api.ds.IitsTagElement;
import de.init.its.ds.ItsAttributes;
import de.init.its.ds.ItsDocument;

/**
 * Service interface for managing the mapping between 
 * LibreoFfice text documents and ITS Tag elements 
 * 
 * @author mkandora
 *
 */
public interface IitsElementManager extends Serializable {
	
	
	/**
	 * UNSAFE OP! REMOVE
	 */
	public void reset();
	
	/**
	 * Inserts an ITS Local Attribute into a document hiearchy
	 * 
	 * @param attrib the {@link ItsAttributes} prefigured constant
	 * @param params ITS Attribute values
	 */
	public IitsTagElement insert(final ItsAttributes attrib, final String...params);
	
	/**
	 * Inserts a new {@link IitsTagElement} using the provided values
	 * 
	 * @param values a mapping between {@link ItsAttributes} and a list of values
	 * @return the newly inserted {@link IitsTagElement}
	 */
	public IitsTagElement insert(final Map<ItsAttributes, List<String>> values);
	
	/**
	 * Inserts an ITS Local Attribute into a document hierarchy
	 * 
	 * @param dcname the name of the data category to insert.
	 * @param dcvalue the value of the data category
	 * @return a new {@link IitsTagElement}
	 */
	public IitsTagElement insert(final String dcname, final String dcvalue);
	
	/**
	 * Inserts an ITS Local Attribute into a document hierarchy using a parent element
	 * for nested data categories
	 * 
	 * @param id the parent element id 
	 * @param dcname the data category nane
	 * @param dcvalue the data category value
	 * @return
	 */
	public IitsTagElement insert(final String id, final String dcname, final String dcvalue);
	
	/**
	 * Inserts an ITS Local Attribute into a document hierarchy using a parent element
	 * for nested data categories
	 * 
	 * @see IitsElementManager#insert(String, String, String)
	 * 
	 * @param id the {@linkplain IitsId}
	 * @param dcname the data category name
	 * @param dcvalue the data category value
	 * @return
	 */
	public IitsTagElement insert(final IitsId id, final ItsAttributes dcname, final String dcvalue);
	
	/**
	 * Removes an {@link IitsTagElement}. The element, if exsisting,
	 * will also be removed from the document tree, resulting all child elements moving 
	 * one depth level higher to the next parent element
	 * 
	 * @return the removed {@linkplain IitsTagElement}
	 */
	public IitsTagElement remove(final String id);
	
	/**
	 * returns an ITS tag element with its id
	 * 
	 * @param id
	 * @return
	 */
	public IitsTagElement getItsElement(final String id);
	
	/**
	 * <p>
	 * Returns all managed {@linkplain IitsTagElement}
	 * </p>
	 * 
	 * @return
	 */
	public List<IitsTagElement> findAll();
	
	/**
	 * returns all its tag elements, that match a certain ITS attribute
	 * @return
	 */
	public List<IitsTagElement> findByAttribute(final IitsAttribute attribute);
	
	
	/**
	 * Returns all Its Tag Elements, that match a certain attribute name
	 * @param attributeName
	 * @return
	 */
	public List<IitsTagElement> findByAttributeName(final String attributeName);
	
	/**
	 * Returns an {@linkplain IitsTagElement} with a given id
	 * 
	 * @param id the Identifier as String
	 * @return the {@link IitsTagElement}
	 */
	public IitsTagElement findById(final String id);
	
	/**
	 * Returns an {@linkplain IitsTagElement} with a given id
	 * 
	 * @param id the Identifier as {@link IitsId}
	 * @return an {@link IitsTagElement}
	 */
	public IitsTagElement findById(final IitsId id);
	
	
	/**
	 * Returns the internally used tag attribute name in
	 * ODT File to resolve ITS Tags
	 * @return
	 */
	public String getIdReferenceAttributeName();
	
	/**
	 * Sets the default tag attribute name for ODT files
	 * to resolve ITS tags  
	 * 
	 * @param name
	 */
	public void setIdReferenceAttributeName(final String name);
	
	/**
	 * Updates the Id
	 * 
	 * @param id
	 */
	public void updateId(final String id) throws Exception;
	
	/**
	 * Sets the currently available {@linkplain ItsDocument}
	 * 
	 * @param document an {@link ItsDocument}
	 */
	public void setItsDocument(final ItsDocument document);
	
	/**
	 * Returns the underlying Its Document
	 * @return {@linkplain ItsDocument}
	 */
	public ItsDocument getItsDocument();
	
	/**
	 * Returns a list of errors
	 * @return
	 */
	public List<ItsError> getErrors();
	
	/**
	 * Returns true, if there were error during its tag processing
	 * @return
	 */
	public boolean hasErrors();
	
	/**
	 * Clears all possible errors
	 */
	public void clearErrors();
}
