 /* ===========================================================================
 *   Copyright (C) 2013 by ]init[ AG für digitale Kommunikation
 * -----------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  See also the full LGPL text here: http://www.gnu.org/copyleft/lesser.html
 * ===========================================================================*/
package de.init.its.process.re;


import java.io.StringReader;
import java.net.URI;
import java.util.HashMap;
import java.util.Map;
import java.util.Stack;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;

import de.init.its.api.ds.IitsTagElement;
import de.init.its.api.process.IitsParserConfig;
import de.init.its.ds.ItsAttributes;
import de.init.its.process.IOResult;
import de.init.its.process.ItsStyleProcessor;
import de.init.its.process.support.Element;
import de.init.its.process.support.ItsTree;
import de.init.its.spi.ItsElementManager;
import de.init.its.util.ItsSerializerService;
import de.init.its.util.Util;

/**
 * Service class to convert between plain ODT documents and ITS annotated ODT Documents.
 * The transformation process retains all available data.
 * 
 * @author m.kandora
 *
 */
public class ItsProcessor {
	
	private XMLReader reader;
	private ODTWriter odtwriter;
	private ItsElementManager manager;
	
	// experimental retrive root after processing
	private ItsTree<Element> tree;
	
	public ItsProcessor() {
		try {
			this.reader = XMLReaderFactory.createXMLReader();
			this.odtwriter = new ODTWriter();
			
			Transformer transform = TransformerFactory.newInstance().newTransformer();
			transform.setParameter(OutputKeys.INDENT, "yes");
			transform.setParameter(OutputKeys.ENCODING, "utf-8");
			transform.setParameter(OutputKeys.METHOD, "xml");
			
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	/**
	 * forward operation to convert odt file from plain to its format
	 *  
	 * @param source the source uri of the odt file
	 * @param target the target to be written
	 * @param manager the (deserialized) manager
	 * @return an IOResult to have a reference to the file
	 */
	public IOResult forward(final URI source, final URI target, final ItsElementManager manager) {
		IOResult result = null;
		try {
			ForwardHandler handler = new ForwardHandler(manager);
			reader.setContentHandler(handler);
			reader.setFeature("http://xml.org/sax/features/namespace-prefixes", true);
			reader.parse(Util.toInputSource(Util.extractZipEntry(source.getPath(), "content.xml")));
			
			StringBuilder builder = new StringBuilder();
			builder.append(IitsParserConfig.XML_DECLARATION);
			builder.append(handler.root.print());
			
			// TODO: write content
			odtwriter.writeXML(source, target, builder.toString(), manager);
			
			
			
			result = new IOResult(target);
			
			// assign tree
			this.tree = handler.root;
			
//			System.out.println(result.getResourcePath());
		} catch(Exception e) {
			e.printStackTrace();
		}
		return result;
	}
	
	/**
	 * backward operation. transforms ITS annotated content to flat enclosing ODT 
	 * Document bookmarks.  
	 * 
	 * @param source the its annotated source document
	 * @param target the target to write the output to
	 * @return
	 */
	public IOResult backward(final URI source, final URI target) {
		IOResult result = null;
		try {
			BackwardHandler handler = new BackwardHandler();
			reader.setContentHandler(handler);
			reader.setFeature("http://xml.org/sax/features/namespace-prefixes", true);
			reader.parse(Util.toInputSource(Util.extractZipEntry(source.getPath(), "content.xml")));
			
			String xmlout = handler.root.print();
			StringBuilder builder = new StringBuilder();
			builder.append(IitsParserConfig.XML_DECLARATION);
			builder.append(xmlout);
			
			// print generated xml
			System.out.println(builder.toString());
			
			manager = handler.manager;
			
			ItsSerializerService.serialize(manager, target.getPath().concat(".its-manager.ser"));
			odtwriter.writeXML(source, target, builder.toString(), manager);
			result = new IOResult(target);
			
			// assign tree
			this.tree = handler.root;
		} catch(Exception e) {
		}
		return result;
	}
	
	/**
	 * backward operation. transforms ITS annotated content to flat enclosing ODT 
	 * Document bookmarks using a preconfigured ITSTree (see also {@linkplain XLIFFProcessor#backward(URI, URI, net.sf.okapi.common.LocaleId, net.sf.okapi.common.LocaleId)}
	 * 
	 * @param source the its annotated source document
	 * @param target the target to write the output to
	 * @return the written odt
	 */
	public IOResult backward(final ItsTree<Element> odtsource, final URI odtstub, final URI target) {
		IOResult result = null;
		try {
			
			BackwardHandler handler = new BackwardHandler();
			reader.setContentHandler(handler);
			reader.setFeature("http://xml.org/sax/features/namespace-prefixes", true);
			reader.parse(new InputSource(new StringReader(odtsource.print())));
			
			String xmlout = handler.root.print();
			StringBuilder builder = new StringBuilder();
			builder.append(IitsParserConfig.XML_DECLARATION);
			builder.append(xmlout);
			
			manager = handler.manager;
			
			ItsSerializerService.serialize(manager, target.getPath().concat(".its-manager.ser"));
			odtwriter.writeXML(odtstub, target, builder.toString(), manager);
			result = new IOResult(target);
			
			// assign tree
			this.tree = handler.root;
			
		} catch(final Exception e) {
			e.printStackTrace();
		}
		
		return result;
	}
	
	/**
	 * Returns the XML parse tree
	 * @return
	 */
	public ItsTree<Element> getTree() {
		return tree;
	}

// ---------------------------------------------------------------------------------------- forward handler
	
	/**
	 * Content handler class for transform plain ODT Documents
	 * to ITS enriched documents by rewriting flat bookmark structures into
	 * possibly deep nested its annotated hierarchies
	 * 
	 * @author mkandora
	 *
	 */
	public static class ForwardHandler extends DefaultHandler
	{
		/** withhelds the characters seen by the parsers */
		private StringBuffer cb;
		
		/** the actual xml tree */
		private ItsTree<Element> root;
		
		/** the next itstree element */
		private ItsTree<Element> next;
		
		/** a working queue */
		private Stack<ItsTree<Element>> stack;
		
		/** the its element manager */
		private ItsElementManager manager;
		
		/**
		 * Creates a new XML Event Handler for a Sax parser
		 * with special regard of bookmark elements, that indicate 
		 * ITS content managed by the ItsElementManager. 
		 * 
		 * The internal implementation uses a lightweight tree 
		 * to maintain the original structure of the document and
		 * to easily insert ITS Data Categories from flat Bookmark elements
		 * to deeply nested ITS annotated Elements. The element by
		 * choice is the text:span taken from the ODF File Specification.
		 * Since, one can assume, that spans do not infer with 
		 * the visible styling of a documents structure, it is 
		 * the optimal choice to embed the Data Categories into spanning 
		 * elements.
		 * 
		 * @param manager the its manager to resolve Data Category Data
		 */
 		public ForwardHandler(final ItsElementManager manager) {
			this.cb = new StringBuffer();
			this.stack = new Stack<ItsTree<Element>>();
			this.manager = manager;
			this.next = null;
		}
		
		@Override
		public void characters(char[] ch, int start, int length)
				throws SAXException {
			if(next == null) next = root;
			next.insert(new Element(cb.append(ch, start, length).toString()));
			cb.setLength(0);
		}
		
		@Override
		public void startElement(String uri, String localName, String qName,
				Attributes attributes) throws SAXException {	
			if(root == null) {
				Map<String, String> namespaces = extractAttributes(attributes, true);
				namespaces.put("its:version", "2.0");
				namespaces.put("xmlns:its", IitsParserConfig.ITS_NAMESPACE_URI);
				root = new ItsTree<Element>(new Element(Util.cut(qName, ":", 0), localName, namespaces, true));
			}
			// only the beginning element is important, since we introduce
			// a deep structure for its data. the end element is not processed, but
			// must be kept from further processing
			else if(isBegin(qName) || isEnd(qName)) 
			{
				if(isBegin(qName)) {
					IitsTagElement itsElement 	= manager.findById(Util.crop(getId(attributes), "[", "]"));
					next = next.insert(createElement(uri, localName, qName, attributes, itsElement));
					stack.push(next);
				}
			}
			else 
			{
				// if next is null, it can only be the parent
				if(next == null) {
					next = root;
				}
				next = next.insert(new Element(Util.cut(qName, ":", 0), localName, extractAttributes(attributes, false), false));
			}
		}
		
		@Override
		public void endElement(String uri, String localName, String qName)
				throws SAXException {
			// drain stack to retrieve the next possible parent element 
			// from hierarchy
			if(!stack.isEmpty()) {
				while(!stack.isEmpty()) next = stack.pop();
			} 
			else {
				if(next != null)
				next = next.getParent(); 
			}
		}
	
		/**
		 * Returns the ITS id
		 * 
		 * @param attributes the elements attributes
		 * @return the id of the its element as string
		 */
		private String getId(Attributes attributes) {
			return attributes.getValue(IitsParserConfig.ODT_FILE_BOOKMARK_REFERENCING_ATTRIBUTE_NAME);
		}
		
		/**
		 * Creates an element
		 * 
		 * @param uri
		 * @param localName
		 * @param qName
		 * @param attributes
		 * @param e
		 * @return
		 */
		public Element createElement(String uri, String localName, String qName, Attributes attributes, IitsTagElement e) {
			// the prefix MUST be a non-conformant ODT element, since it is needed to know exactly 
			// when ITS content is being seen in the retransformation process
			// also see backwardhandler
			return new Element("its", "span", e.renderAttributes());
		}
		
		/**
		 * TODO pull up to abstract class
		 * 
		 * Service routine to extract the attributes from a given element
		 * 
		 * @param attributes
		 * @return
		 */
		private Map<String, String> extractAttributes(Attributes attributes, boolean extractNS) {
			if(extractNS) return extractNamespaces(attributes);
			Map<String, String> map = new HashMap<String, String>();
			for(int i = 0; i < attributes.getLength(); i++) {
				map.put(attributes.getQName(i), attributes.getValue(i));
			}
			return map;
		}
		
		/**
		 * TODO Pull up to abstract class
		 * 
		 * Service routine to extract the namespaces of an element. 
		 * Only useful for root elements, but can be applied to any element
		 * inside the document tree
		 * 
		 * @param attributes
		 * @return
		 */
		private Map<String, String> extractNamespaces(Attributes attributes) {
			Map<String, String> map = new HashMap<String, String>();
			for(int i = 0; i < attributes.getLength(); i++) {
				if(attributes.getQName(i).startsWith("xmlns"))
				map.put(attributes.getQName(i), attributes.getValue(i));
			}
			return map;
		}
		
		/**
		 * Returns true, if element is bookmark begin
		 * 
		 * @param qname the name of the element
		 * @return
		 */
		private boolean isBegin(String qname) {
			return IitsParserConfig.ODT_FILE_BOOKMARK_BEGIN_ELEMENT_NAME_NS.equals(qname);
		}
		
		/**
		 * Returns true, if element is bookmark end
		 * 
		 * @param qname the name of the element
		 * @return
		 */
		private boolean isEnd(String qname) {
			return IitsParserConfig.ODT_FILE_BOOKMARK_END_ELEMENT_NAME_NS.equals(qname);
		}
		
	}
	
	
// ---------------------------------------------------------------------------------------- backward handler
	/**
	 * Content handler to transform ODT Documents with 
	 * ITS annotated content back to plain ODT Documents with bookmarks as
	 * placeholders for its data
	 * 
	 * @author mkandora
	 *
	 */
	public static class BackwardHandler extends DefaultHandler
	{
		private static final String ITS_ELEMENT_NAME = "its:span";
		
		private IitsTagElement itsroot, itsnext;
		private ItsTree<Element> root, next;
		private StringBuffer sb;
		
		private ItsElementManager manager;
		
		/**
		 * Creates a new content handler to transform ITS annotated
		 * ODT documents back to plain ODT document format. Each
		 * nested element is converted to a flat bookmark. the structure
		 * is retained except for the deep its annotated content. 
		 * while parsing the file, the manager is being created to 
		 * preserve the id information to handle the its content
		 */
		public BackwardHandler() {
			sb 		= new StringBuffer();
			manager = ItsElementManager.getManager(true);
		}
		
		@Override
		public void characters(char[] ch, int start, int length)
				throws SAXException {
			if(next == null) next = root;
			next.insert(new Element(sb.append(ch, start, length) . toString()));
			sb.setLength(0);
		}
		
		@Override
		public void startElement(String uri, String localName, String qName,
				Attributes attributes) throws SAXException {
			Map<String, String> map = new HashMap<String, String>();
			
			if(root == null) {
				Map<String, String> namespaces = extractAttributes(attributes, true);
				
				// delegate?
				namespaces.put(ITS_ELEMENT_NAME, IitsParserConfig.ITS_NAMESPACE_URI);
				root = new ItsTree<Element>(new Element(Util.cut(qName, ":", 0), localName, namespaces, true));
			}
			else if(isItsElement(qName)) {
				if(next == null) {
					next = root;
				}
				map.clear();
				
				// TODO: create new Element
				putItsElement(itsnext, extractAttributes(attributes, false));
				
				if(itsnext != null) 
				{ 
					// added styling
					next = next.insert(new Element("text", "span", "text:style-name", ItsStyleProcessor.getProcessor().getStyleId(itsnext)));
					
					map.put(IitsParserConfig.ODT_FILE_BOOKMARK_REFERENCING_ATTRIBUTE_NAME, "its-[" + itsnext.getId().getString() + "]");
					next = next.insert(new Element(IitsParserConfig.ODT_FILE_BOOKMARK_ELEMENT_NS_PREFIX, IitsParserConfig.ODT_FILE_BOOKMARK_BEGIN_ELEMENT_NAME, map, false)).getParent();
				}
			}
			else {
				if(next == null) {
					next = root;
				}
				// normal element
				next = next.insert(new Element(Util.cut(qName, ":", 0), localName, extractAttributes(attributes, false), false));
			}
		}
		
		@Override
		public void endElement(String uri, String localName, String qName)
				throws SAXException {
			
			if(isItsElement(qName)) {
				Map<String, String> map = new HashMap<String, String>();
				// TODO: create new Element
				
				map.put(IitsParserConfig.ODT_FILE_BOOKMARK_REFERENCING_ATTRIBUTE_NAME, "its-[" + itsnext.getId().getString() + "]");
				
				
				// MK: one level higher in hierarchy
				next = next.insert(new Element(
						IitsParserConfig.ODT_FILE_BOOKMARK_ELEMENT_NS_PREFIX, IitsParserConfig.ODT_FILE_BOOKMARK_END_ELEMENT_NAME, map, false)).getParent().getParent();
				
				// move up its element
				if(itsnext.getParentElement().equals(itsroot))
				itsnext = itsroot; 
				else itsnext = itsnext.getParentElement();
				
			}
			else {
				if(next != null)
				next = next.getParent(); 
			}
		}
		
		/**
		 * returns true, if it is an ITS element
		 * 
		 * <b>NOTICE: </b>this forward transformation explicitly sets the surrounding element
		 * as its:span, not text:span, using the its prefix to indicate ITS content. 
		 * otherwise, it would be rather impossible to detect its annotated content, since
		 * text:span is ambigious. this approach has no impact since the generated ODT
		 * document is only used as intermediary format to export and import to XLIFF
		 * 
		 * @param qname
		 * @return
		 */
		protected boolean isItsElement(final String qname) {
			return ITS_ELEMENT_NAME.equals(qname);
		}
		
		/**
		 * TODO Pull up to abstract class
		 * 
		 * Service routine to extract the attributes from a given element
		 * 
		 * @param attributes
		 * @return
		 */
		private Map<String, String> extractAttributes(Attributes attributes, boolean extractNS) {
			if(extractNS) return extractNamespaces(attributes);
			Map<String, String> map = new HashMap<String, String>();
			for(int i = 0; i < attributes.getLength(); i++) {
				map.put(attributes.getQName(i), attributes.getValue(i));
			}
			return map;
		}
		
		/**
		 * TODO Pull up to abstract class
		 * 
		 * service routine to extract the namespaces of a given element. this might only
		 * be useful for root elements
		 */
		private Map<String, String> extractNamespaces(Attributes attributes) {
			Map<String, String> map = new HashMap<String, String>();
			for(int i = 0; i < attributes.getLength(); i++) {
				if(attributes.getQName(i).startsWith("xmlns"))
				map.put(attributes.getQName(i), attributes.getValue(i));
			}
			return map;
		}
		
		/**
		 * Creates a new its element preserving the hierarchical structure 
		 * 
		 * @param map
		 * @return
		 */
		protected void putItsElement(final IitsTagElement parent, final Map<String, String> map) {
			IitsTagElement element = null;
			for(String nskey : map.keySet()) {
				ItsAttributes its = ItsAttributes.valueOf(Util.cut(nskey, ":", 1));
				if(element == null && parent == null) {
					element = manager.insert(its, map.get(nskey));
					itsroot = element;
					continue;
				}
				else if(element == null) {
					// insert as child
					element = manager.insert(parent.getId(), its, map.get(nskey));
					continue;
				}
				element.addAttribute(its, map.get(nskey));
			}
			itsnext = element;
		}
		
	}

	
	
}
