 /* ===========================================================================
 *   Copyright (C) 2013 by ]init[ AG für digitale Kommunikation
 * -----------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  See also the full LGPL text here: http://www.gnu.org/copyleft/lesser.html
 * ===========================================================================*/
package de.init.its.process.re;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.net.URI;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Node;

import de.init.its.process.ItsODTDocumentType;
import de.init.its.spi.ItsElementManager;
import de.init.its.util.Util;

/**
 * Service class for writing ODT Files
 * 
 * @author m.kandora
 *
 */
public class ODTWriter {

	private Transformer transformer;
	
	public ODTWriter() {
		try {
		this.transformer 	= TransformerFactory.newInstance().newTransformer();
		} catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Writes XML to original ODT File. Merges the XML content into content.xml
	 * 
	 * @param source the source file
	 * @param target the target to be written
	 * @param xml the xml source as string
	 * @param mgr the mgr
	 * @throws Exception
	 */
	public void writeXML(final URI source, final URI target, final String xml, final ItsElementManager mgr) throws Exception
	{	
		File content = File.createTempFile("content", ".xml");
		Util.writeToDisk(content.getAbsolutePath(), xml);
		ODTFileWriter zipWriter = new ODTFileWriter();
		
		zipWriter.beginZip(new File(target.getPath()));
		
		ZipInputStream zis = new ZipInputStream(new FileInputStream(source.getPath()));
		for(ZipEntry next = zis.getNextEntry(); next != null; next = zis.getNextEntry()) {
			if("content.xml".equals(next.getName())) {
				
				ZipEntry contentXML = new ZipEntry("content.xml");
				
				// determine file size
				FileInputStream fis = new FileInputStream(content);
				contentXML.setSize(fis.getChannel().size());
				fis.close();
				
				zipWriter.replace(contentXML, content);
			} else {
				try {
					zipWriter.insert(next, zis);
					
				} catch (Exception e) {
					// just skip
					e.printStackTrace();
				}
			}
		}
		zis.close();
		zipWriter.endZip();
	}
	
	/**
	 * Write XML to disk using dom documents
	 * 
	 * @param rootDocument the processed xml root element
	 * @param target the target location
	 * @throws IOException 
	 * @throws TransformerException
	 * @deprecated
	 */
	public void writeXML(final Node rootDocument, final String target, final ItsODTDocumentType type, final String origResource, final ItsElementManager mgr) throws IOException,
			TransformerException {
		this.transformer.setOutputProperty("indent", "true");
		boolean mgrInserted = false;
		// last step: write out ODT Document
		// write flat
		if(ItsODTDocumentType.FLATODT.equals(type)) {
			File f = new File(target);
			if(!f.exists()) f.createNewFile();
			this.transformer.transform(new DOMSource(rootDocument), new StreamResult(f));
		} else { // write default zipped odt
			File f = File.createTempFile("content", "xml");
			if(f.exists()) {
				ZipInputStream zis = new ZipInputStream(new FileInputStream(origResource));
				try {
					String extension = ".odt";
					File tempZip = File.createTempFile("tmp_odt", extension);
					File tempContent = File.createTempFile("content", ".xml");
					
					// prepare content.xml
					this.transformer.transform(new DOMSource(rootDocument), new StreamResult(tempContent));
					
					ODTFileWriter zipWriter = new ODTFileWriter();
					zipWriter.beginZip(tempZip);
					
					for(ZipEntry next = zis.getNextEntry(); next != null; next = zis.getNextEntry()) {
//						ZipEntry next = enumeration.nextElement();
						if("content.xml".equals(next.getName())) {
							
							ZipEntry contentXML = new ZipEntry("content.xml");
							
							// ugly hack to determine file size
							FileInputStream fis = new FileInputStream(tempContent);
							contentXML.setSize(fis.getChannel().size());
							fis.close();
							
							zipWriter.replace(contentXML, tempContent);
						} else {
							try {
								
								zipWriter.insert(next, zis);
								
								if(mgr != null && !mgrInserted) {
									
									ZipEntry serial = new ZipEntry("its.ser");
									ByteArrayOutputStream baos = new ByteArrayOutputStream();
									ObjectOutputStream oos = new ObjectOutputStream(baos);
									oos.writeObject(mgr);
									oos.flush();
									oos.close();
									ZipInputStream serialZIS = new ZipInputStream(new ByteArrayInputStream(baos.toByteArray()));
									zipWriter.insert(serial, serialZIS);
									mgrInserted = true;
								}
								
							} catch (Exception e) {
								// just skip
								e.printStackTrace();
								
							}
						}
					}
					zis.close();
					zipWriter.endZip();
					
					// FIXME: but it's a blocking resource and cannot be read
					// UPDATE: no acces from UNO context. this may be used only 
					// in conjunction with an external tool set to process and existing
					// plain ODT resource
					File orig = new File(origResource);
//					String path = orig.getAbsolutePath();
					/*boolean deleteSuccessFul = */
					orig.delete();
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}
	}
	
	/**
	 * Service class to write ODT Document Files. 
	 * 
	 * @author m.kandora
	 *
	 */
	protected static class ODTFileWriter
	{
		private ZipOutputStream zos;
		
		public void insert(final ZipEntry entry, final ZipInputStream zis) throws Exception {
			zos.putNextEntry(entry);
			
			byte[] b = new byte[0x1000];
			int len = 0;
			while((len = zis.read(b)) != -1) {
				zos.write(b, 0, len);
			}
			
			zos.closeEntry();
		}
		
		public boolean replace(final ZipEntry entry, final File f) throws Exception {
			
			zos.putNextEntry(entry);
			
			FileInputStream fis = new FileInputStream(f);
			byte[] buffer = new byte[0x1000];
			int len = -1;
			while((len = fis.read(buffer)) != -1) {
				zos.write(buffer, 0, len);
			}
			fis.close();
			zos.flush();
			return false;
		}
		
		public void beginZip(final File target) throws Exception {
			zos = new ZipOutputStream(new FileOutputStream(target));
		}
		
		public void endZip() throws Exception {
			zos.flush();
			zos.close();
		}
	}
}
