 /* ===========================================================================
 *   Copyright (C) 2013 by ]init[ AG für digitale Kommunikation
 * -----------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  See also the full LGPL text here: http://www.gnu.org/copyleft/lesser.html
 * ===========================================================================*/
package de.init.its.process.re;

import java.io.StringReader;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;

import de.init.its.ds.ItsAttributes;
import de.init.its.ds.ItsAttributes.DataCategories;
import de.init.its.process.support.Element;
import de.init.its.process.support.ItsTree;
import de.init.its.util.Util;

/**
 * A Simple XLIFF Reader. This reader will convert an XLIFF Source
 * into a tree format something like this:
 * 
 * 								XLIFFDocument
 * 									|
 * 									File
 * 									|
 * 									Body
 * 							/		|		\
 * 							TU		TU		TU
 * 						/		\ /		\ /		\
 * 						S		T S		T S		T
 * 					/	|	\
 * 				g		g	g
 * 
 * where TU is a translation unit consisting of (S)ource and (T)arget elements.
 * Each Source and Target element can have text nodes, simple generic xliff groupings 
 * and the relevant MRK Elements indicating its content. 
 * 
 * This XLIFF Reader is not general purpose! Instead it may be only useful in this
 * context
 * 
 * @author m.kandora
 *
 */
public class XLIFFReader {
	
	private XMLReader reader;
	private XLIFFHandler handler;
	
	public XLIFFReader() { 
		try {
			this.reader = XMLReaderFactory.createXMLReader();
			this.reader.setFeature("http://xml.org/sax/features/namespace-prefixes", true);
			this.handler = new XLIFFHandler();
			this.reader.setContentHandler(handler);
			
		} catch(final Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Read an xml as String and return it as {@linkplain ItsTree} represenation
	 * @param xml
	 * @return
	 */
	public ItsTree<Element> read(final String xml) {
		try {
			// phase1: parse the xliff as string. the tree will be stored internally...
			this.reader.parse(new InputSource(new StringReader(xml)));
			
			// .. and returned here for later processing
			return this.handler.root;
		} catch(final Exception e) {
			e.printStackTrace();
		}
		return new ItsTree<Element>(new Element(null));
	}
	
	/**
	 * Read an external XLIFF Source and return a {@linkplain ItsTree} representation of this source
	 * @param xliffsource
	 * @return
	 */
	public ItsTree<Element> read(final URI xliffsource) {
		try {
			// phase1: parse the xliff file. the tree will be stored internally...
			this.reader.parse(Util.toInputSource(xliffsource.getPath()));
			
			// .. and returned here for later processing
			return this.handler.root;
		} catch(Exception e) {
			e.printStackTrace();
		}
		// do not return null, instead deliver a tree with a single text node
		// containing a null element
		return new ItsTree<Element>(new Element(null));
	}
	
	/**
	 * XLIFF to ITS Mapper Class. 
	 * This service class provides a bunch of method to detect ITS annotated 
	 * content in an XLIFF File. 
	 * 
	 * @author mkandora
	 *
	 */
	public static class XLIFF2ITSMapperService {
		
		/**
		 * the 'extended' xliff its prefix
		 */
		final static String prefixExt = "itsxlf";
		
		/**
		 * the 'normal' its prefix 
		 */
		final static String prefixNor = "its";
		
		/**
		 * Returns true, if subtree elements head is a localization note
		 * @param t the subtree
		 * @return true, if localization note dc
		 */
		public static boolean isLocNote(final ItsTree<Element> t) {
			return Util.matchesAny(t.getHead().getAttributes(), prefixExt + ":locNoteType", "comment") || 
					Util.matchesAny(t.getHead().getAttributes(), prefixNor + ":locNoteType", prefixNor + ":locNote");
		}
		
		/**
		 * Returns true, if subtree elements head is a translate dc
		 * @param t the subtree
		 * @return true, if translate dc
		 */
		public static boolean isTranslate(final ItsTree<Element> t) {
			return Util.matchesAny(t.getHead().getAttributes(), "") 
					|| t.getHead().getAttributes().containsKey("mtype") ? 
							t.getHead().getAttributes().get("mtype").equals("protected") : false; 
		}
		
		/**
		 * Returns true, if subtree elements head is a localization filter
		 * @param t the subtree
		 * @return true, if localization filter dc
		 */
		public static boolean isLocFilter(final ItsTree<Element> t) {
			return Util.matchesAny(t.getHead().getAttributes(), prefixNor + ":localeFilterList", prefixNor + ":localeFilterType");
		}

		/**
		 * Returns true, if subtree elements head is a term
		 * @param t the subtree
		 * @return true, if term dc
		 */
		public static boolean isTerminology(final ItsTree<Element> t) {
			return Util.matchesAny(t.getHead().getAttributes(), prefixExt + ":termConfidence", prefixExt + ":termInfo")
					|| t.getHead().getAttributes().containsKey("mtype") ? 
							t.getHead().getAttributes().get("mtype").equals("term") : false
					|| Util.matchesAny(t.getHead().getAttributes(),  prefixNor + ":termConfidence", prefixNor + "termInfo");
		}
		
		/**
		 * Tries to detect a certain data category
		 * 
		 * @param t the tree, that possibly contains an ITS data category
		 * @return the detected {@linkplain DataCategories} or <b>null</b>
		 */
		public static DataCategories detect(final ItsTree<Element> t) {
			if(t.getHead() == null) return null;
			if(isLocFilter(t)) return DataCategories.LOCALEFILTER;
			if(isLocNote(t)) return DataCategories.LOCALIZATIONNOTE;
			if(isTerminology(t)) {
				return DataCategories.TERMINOLOGY;
			}
			if(isTranslate(t)) return DataCategories.TRANSLATE;
			return null;
		}
		
		/**
		 * Returns a mapping for the respective data category and the XLIFF mapping
		 * The mapping returned by this method refers to this document:
		 * <p>
		 * <a href="http://www.w3.org/International/its/wiki/XLIFF_1.2_Mapping#Localization_Note">http://www.w3.org/International/its/wiki/XLIFF_1.2_Mapping#Localization_Note</a>
		 * </p>
		 * @param t the current its annotated tree element to convert
		 * @param dc the infered data category
		 * @return the mapping for its attributes
		 */
		public static Map<ItsAttributes, List<String>> convert(final ItsTree<Element> t, final DataCategories dc) {
			Map<ItsAttributes, List<String>> mapping = new HashMap<ItsAttributes, List<String>>();
			Element head = t.getHead();
			Map<String, String> attributes = head.getAttributes();
			List<String> list = null;
			switch(dc) {
			case LOCALEFILTER:
				for(String key : attributes.keySet()) {
					if((prefixNor + ":localeFilterList").equals(key)) {
						if(mapping.containsKey(ItsAttributes.localeFilterList))
						{
							list = mapping.get(ItsAttributes.localeFilterList);
						} 
						else {
							list = new ArrayList<String>();
						}
						list.add(attributes.get(key));
						mapping.put(ItsAttributes.localeFilterList, list);
					}
					else if((prefixNor + ":localeFilterType").equals(key)) {
						if(mapping.containsKey(ItsAttributes.localeFilterType)) {
							list = mapping.get(ItsAttributes.localeFilterType);
						} else  {
							list = new ArrayList<String>();
						}
						list.add(attributes.get(key));
						mapping.put(ItsAttributes.localeFilterType, list);
					}
				}
				break;
			case LOCALIZATIONNOTE:
				for(String key : attributes.keySet()) {
					if((prefixExt + ":locNoteType").equals(key)) {
						if(mapping.containsKey(ItsAttributes.locNoteType)) {
							list = mapping.get(ItsAttributes.locNoteType);
						} else {
							list = new ArrayList<String>();
						}
						list.add(attributes.get(key));
						mapping.put(ItsAttributes.locNoteType, list);
					}
					if("comment".equals(key)) {
						if(mapping.containsKey(ItsAttributes.locNote)) {
							list = mapping.get(key);
						} else {
							list = new ArrayList<String>();
						}
						list.add(attributes.get(key));
						mapping.put(ItsAttributes.locNote, list);
					}
					
					// TODO: Missing. No explizit mapping for LocNoteRef on
					// http://www.w3.org/International/its/wiki/XLIFF_1.2_Mapping#Localization_Note
				}
				break;
			case TERMINOLOGY:
				for(String key : attributes.keySet()) {
					if("mtype".equals( key ) && attributes.get(key).equals("term")) {
						
						if(mapping.containsKey(ItsAttributes.term)) {
							list = mapping.get(ItsAttributes.term);
						} else {
							list = new ArrayList<String>();
						}
						list.add("yes");
						mapping.put(ItsAttributes.term, list);
					}
					else if((prefixExt + ":termConfidence").equals(key)) {
						if(mapping.containsKey(ItsAttributes.termConfidence)) {
							list = mapping.get(ItsAttributes.termConfidence);
						}
						else {
							list = new ArrayList<String>();
						}
						list.add(attributes.get(key));
						mapping.put(ItsAttributes.termConfidence, list);
					}
					else if((prefixExt + ":termInfo").equals(key)) {
						if(mapping.containsKey(ItsAttributes.termInfoRef)) {
							list = mapping.get(ItsAttributes.termInfoRef);
						} else {
							list = new ArrayList<String>();
						}
						list.add(attributes.get(key));
						mapping.put(ItsAttributes.termInfoRef, list);
					}
				}
				break;
			case TRANSLATE:
				for(String key : attributes.keySet()) {
					if("mtype".equals(key) && attributes.get(key).equals("protected")) {
						// translate = yes is getting lost during the conversion process
						// but does it matter? if everything is translatable, a semantic loss
						// is not given...
						if(mapping.containsKey(ItsAttributes.translate)) {
							list = mapping.get(ItsAttributes.translate);
						} else {
							list = new ArrayList<String>();
						}
						list.add("no");
						mapping.put(ItsAttributes.translate, list);
					}
				}
				break;
			default:
				break;
			}
			return mapping;
		}
	}
	
	/**
	 * Simple xliff XML Content Handler
	 */
	static class XLIFFHandler extends DefaultHandler {
		
		private ItsTree<Element> root, next;
		private StringBuilder cb;
		private Stack<ItsTree<Element>> stack;
		
		
		public XLIFFHandler() {
			this.cb 		= new StringBuilder();
			this.stack 		= new Stack<ItsTree<Element>>();
		}
		
		@Override
		public void characters(char[] ch, int start, int length)
				throws SAXException {
			if(next == null) next = root;
			cb.append(ch, start, length);
			
			// skip whitespace
			if(cb.toString().matches("[\r\n]")) {
				cb.setLength(0);
				return;
			}
			
			next.insert(te(cb.toString()));
			cb.setLength(0);
		}
		
		@Override
		public void startElement(String uri, String localName, String qName,
				Attributes attributes) throws SAXException {
			if(root == null) { // root case 
				root = new ItsTree<Element>(ro(pre(qName), localName, attributes));
			}
			else { // default case
				if(next == null) next = root;
				next = next.insert(el(pre(qName), localName, attributes));
			}
 		}
		
		@Override
		public void endElement(String uri, String localName, String qName)
				throws SAXException {
			// drain stack
			if(!stack.isEmpty()) {
				while(!stack.isEmpty()) next = stack.pop();
			} 
			// retrieve parent element and go up one level in hierarchy
			else if(next != null) next = next.getParent();
		}
		
		/**
		 * Creates a new text node element and returns it
		 * @param text
		 * @return a text node as element
		 */
		protected Element te(final String text) {
			return new Element(text);
		}
		
		/**
		 * Creates a new element with given prefix, name and optional attributes
		 * @param prefix
		 * @param name
		 * @param attributes
		 * @return an element node
		 */
		protected Element el(final String prefix, final String name, final Attributes attributes) {
			return new Element(prefix, name, extractAttributes(attributes, false), false);
		}
		
		/**
		 * creates a new root element
		 * 
		 * @param prefix
		 * @param name
		 * @param attributes
		 * @return the root node as element
		 */
		protected Element ro(final String prefix, final String name, final Attributes attributes) {
			Element element = new Element(prefix, name, extractNamespaces(attributes), true);
			element.addAttributes(extractAttributes(attributes, false));
			return element;
		}
		
		/**
		 * returns the prefix of a full qualified named element
		 * @param qName the qualified named element name
		 * @return the prefix
		 */
		protected String pre(final String qName) {
			String prefix = Util.cut(qName, ":", 0);
			// if prefix is same as qname, qname has no prefix
			return qName.equals(prefix) ? null : prefix;
		}
		
		/**
		 * Service routine to extract the attributes from a given element
		 * 
		 * @param attributes
		 * @return
		 */
		private Map<String, String> extractAttributes(Attributes attributes, boolean extractNS) {
			if(extractNS) return extractNamespaces(attributes);
			Map<String, String> map = new HashMap<String, String>();
			for(int i = 0; i < attributes.getLength(); i++) {
				if(!attributes.getQName(i).startsWith("xmlns"))
				map.put(attributes.getQName(i), attributes.getValue(i));
			}
			return map;
		}
		
		/**
		 * Service routine to extract the namespaces of an element. 
		 * Only useful for root elements, but can be applied to any element
		 * inside the document tree
		 * 
		 * @param attributes
		 * @return
		 */
		private Map<String, String> extractNamespaces(Attributes attributes) {
			Map<String, String> map = new HashMap<String, String>();
			for(int i = 0; i < attributes.getLength(); i++) {
				if(attributes.getQName(i).startsWith("xmlns"))
				map.put(attributes.getQName(i), attributes.getValue(i));
			}
			return map;
		}
	}

}
