 /* ===========================================================================
 *   Copyright (C) 2013 by ]init[ AG für digitale Kommunikation
 * -----------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  See also the full LGPL text here: http://www.gnu.org/copyleft/lesser.html
 * ===========================================================================*/
package de.init.its.process.support;

import java.net.URI;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.init.its.ds.ItsAttributes;

/**
 * A Simple xml element
 * 
 * @author m.kandora
 *
 */
public class Element {
	
	private String name;
	private Map<String, String> attributes;
	private Map<String, URI> namespaces;
	private String prefix;
	private String text;
	
	public Element(final String text) {
		this.text = text;
		this.attributes = new HashMap<String, String>();
	}
	
	public Element(final String prefix, final String name, String...attr) {
		if(name.matches("\\s+")) {
			System.err.println("Element Names SHALL NOT contain whitespaces");
		}
		this.name = name;
		this.prefix = prefix != null ? (prefix.isEmpty() ? null : prefix) : null;
		this.attributes = new HashMap<String, String>();
		this.namespaces = new HashMap<String, URI>();
		
		if((attr.length & 1) == 0) {
			int index = 0;
			String key = "";
			for(String val : attr) {
				if(index % 2 == 0 ) {
					key = val;
				} else {
					 attributes.put(key, val);
				}
				index++;
			}
		} else {
			System.err.println("ERROR!! Illegal attribute count! Must be even!");
		}
	}
	/**
	 * Creates a simple element
	 * @param prefix
	 * @param name
	 * @param attr
	 */
	public Element(final String prefix, final String name, final Map<ItsAttributes, List<String>> attr) {
		this.name = name;
		this.prefix = prefix;
		this.attributes = new HashMap<String, String>();
		this.namespaces = new HashMap<String, URI>();
		
		for(ItsAttributes its : attr.keySet()) {
			attributes.put("its:" + its.name(), attr.get(its).toString().replaceAll("(\\[|\\])", ""));
		}
	}
	
	public Element(final String prefix, final String name, final Map<String, String> attr, boolean nsonly) {
		this.name = name;
		this.prefix = prefix;
		this.attributes = new HashMap<String, String>();
		this.namespaces = new HashMap<String, URI>();
		
		// we need to check for namespaces directly
		for(final String key : attr.keySet()) {
			if(key.startsWith("xmlns")) {
				namespaces.put(key, URI.create(attr.get(key)));
			} else if(!nsonly){
				this.attributes.putAll(attr);
				break;
			}
		}
	}
	
	public void putNameSpace(final String prefix, final URI namespace) {
		this.namespaces.put(prefix, namespace);
	}
	
	public void putNameSpace(final String prefix, final String namespace) {
		putNameSpace(prefix, URI.create(namespace));
	}
	
	public void putAttribute(final String key, final String value) {
		this.attributes.put(key, value);
	}
	
	public void addAttributes(final Map<String, String> map) {
		this.attributes.putAll(map);
	}
	
	/**
	 * returns the namespaces
	 * 
	 * @return
	 */
	public Map<String, URI> getNameSpaces() {
		return namespaces;
	}
	
	/**
	 * Returns the attributes
	 * @return
	 */
	public Map<String, String> getAttributes() {
		return attributes;
	}
	
	/**
	 * Returns the name if this element
	 * @return
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * Returns the namespace prefix of this element, if any
	 * @return
	 */
	public String getPrefix() {
		return prefix;
	}
	/**
	 * renders the start of an element
	 * 
	 * @param empty, if element is empty
	 * @return
	 */
	protected String start(final boolean empty) {
		StringBuilder b = new StringBuilder();
		b.append("<");
		b.append(prefix != null ? prefix + ":" : "");
		
		b.append(name);
		for(String key : attributes.keySet()) {
			b.append(" ");
			b.append(key);
			b.append("=");
			b.append("\"");
			b.append(attributes.get(key));
			b.append("\"");
		}
		
		for(String ns : namespaces.keySet()) {
			b.append(" ");
//			b.append("xmlns:");
			b.append(ns);
			b.append("=");
			b.append("\"");
			b.append(namespaces.get(ns));
			b.append("\"");
		}
		b.append(empty ? "/" : "");
		b.append(">");
		return b.toString();
	}
	
	protected String end() {
		return "</" + (prefix != null ? (prefix + ":") : "") + name + ">" ;
	}
	
	/**
	 * returns the text from this text node. check with {@link #isText()} first
	 * 
	 * @return
	 */
	public String getText() {
		return text;
	}
	
	/**
	 * Returns true, if this element is a text node
	 * @return
	 */
	public boolean isText() {
		return text != null;
	}

	@Override
	public String toString() {
		return "Element [name=" + name + ", attributes=" + attributes
				+ ", namespaces=" + namespaces + ", prefix=" + prefix
				+ ", text=" + text + "]";
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((attributes == null) ? 0 : attributes.hashCode());
		result = prime * result + ((name == null) ? 0 : name.hashCode());
		result = prime * result
				+ ((namespaces == null) ? 0 : namespaces.hashCode());
		result = prime * result + ((prefix == null) ? 0 : prefix.hashCode());
		result = prime * result + ((text == null) ? 0 : text.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Element other = (Element) obj;
		if (attributes == null) {
			if (other.attributes != null)
				return false;
		} else if (!attributes.equals(other.attributes))
			return false;
		if (name == null) {
			if (other.name != null)
				return false;
		} else if (!name.equals(other.name))
			return false;
		if (namespaces == null) {
			if (other.namespaces != null)
				return false;
		} else if (!namespaces.equals(other.namespaces))
			return false;
		if (prefix == null) {
			if (other.prefix != null)
				return false;
		} else if (!prefix.equals(other.prefix))
			return false;
		if (text == null) {
			if (other.text != null)
				return false;
		} else if (!text.equals(other.text))
			return false;
		return true;
	}
	
	
}