 /* ===========================================================================
 *   Copyright (C) 2013 by ]init[ AG für digitale Kommunikation
 * -----------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  See also the full LGPL text here: http://www.gnu.org/copyleft/lesser.html
 * ===========================================================================*/
package de.init.its.process.support;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Deque;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * tree data structure implementation
 * 
 * @author m.kandora
 *
 * @param <T>
 */
public class ItsTree<T extends Element> {
	
	protected ItsTree<T> parent, next;
	protected T head;
	protected List<ItsTree<T>> children;
	protected Map<T, ItsTree<T>> mapping = new HashMap<T, ItsTree<T>>();
	
	protected static final String EOL = System.getProperty("line.separator");
	
	/**
	 * Creates a new Tree with an head element
	 * @param head
	 */
	public ItsTree(final T head) {
		this.head 		= head;
		this.children 	= new ArrayList<ItsTree<T>>();
		this.mapping.put(head, this);
	}
	
	/**
	 * Inserts an element as child for the parent and returns the child subtree
	 * @param parent
	 * @param node
	 * @return
	 */
	public ItsTree<T> insert(final T parent, final T node) {
		if(mapping.containsKey(parent)) {
			return mapping.get(parent).insert(node);
		} else {
			return insert(parent).insert(node);
		}
	}
	
	/**
	 * Inserts an element as child to this tree and returns the child subtree
	 * 
	 * @param node
	 * @return
	 */
	public ItsTree<T> insert(final T node) {
		return insert(new ItsTree<T>(node));
	}
	
	/**
	 * Inserts a subtree as a child to this node
	 * @param tree the tree to insert at this node
	 * @return the inserted tree (is equal to the parameter, useful for method chaining)
	 */
	public ItsTree<T> insert(final ItsTree<T> tree) {
		if(hasChildren()) children.get(children.size() - 1).next = tree;
		children.add(tree);
		tree.parent = this;
		tree.mapping = this.mapping;
		mapping.put(tree.getHead(), tree);
		return tree;
	}
	
	/**
	 * Removes a node form the tree and returns the subtree
	 * @param node
	 * @return
	 */
	public ItsTree<T> remove(final T node) {
		if(mapping.containsKey(node)) {
			ItsTree<T> subtree = mapping.remove(node);
			if(subtree.parent != null)
				subtree.parent.children.remove(subtree);
			return subtree;
		}
		return null;
	}
	
	/**
	 * Returns the head of this node
	 * @return
	 */
	public T getHead() {
		return this.head;
	}
	
	/**
	 * Returns the parent of this node
	 * @return
	 */
	public ItsTree<T> getParent() {
		return parent;
	}
	
	/**
	 * Returns the children of this node
	 * @return
	 */
	public List<ItsTree<T>> getChildren() {
		return children;
	}
	
	/**
	 * Returns true, if this subtree contains children
	 * @return
	 */
	public boolean hasChildren() {
		return !children.isEmpty();
	}
	
	/**
	 * returns the successive elements of this node
	 * @return
	 */
	public Collection<T> getSuccessors() {
		return getSuccessors(head);
	}
	
	/**
	 * returns the successors ofthe root element
	 * @param root
	 * @return
	 */
	public Collection<T> getSuccessors(final T root) {
		List<T> successors = new ArrayList<T>();
		for(ItsTree<T> child : mapping.get(root).children) {
			successors.add(child.getHead());
		}
		return successors;
	}
	
	/**
	 * Returns true, if the head of this tree is empty (eg. null)
	 * @return true, if head is null
	 */
	public boolean isEmpty() {
		return this.head == null;
	}
	
	/**
	 * Returns a list of all Elements specified by name an prefix.
	 * An empty prefix will be interpreted as null by Element.
	 *
	 * @param prefix the namespace prefix of the elements to look for
	 * @param name the name of the elements to look for
	 * @return a deque with all subtrees containing this element to retain the structural data
	 */
	public Deque<ItsTree<T>> findAllByName(final String prefix, final String name) {
		Deque<ItsTree<T>> coll = new LinkedList<ItsTree<T>>();
		for(T child : mapping.keySet()) {
			if(child.getName() == null) continue;
			if(child.getName().equals(name)) {
				if(prefix == null) coll.addFirst(mapping.get(child));
				else if(prefix.equals(child.getPrefix())) coll.addFirst(mapping.get(child));
			}
		}
		return coll;
	}
	/**
	 * Returns the first Element, that is matched by prefix and the name
	 * @param prefix the namespace prefix of the Element
	 * @param name the elements name
	 * @return the first matching subtree containg the desired element
	 */
	public ItsTree<T> findByName(final String prefix, final String name) {
		ItsTree<T> sub = null;
		if((sub = mapping.get(new Element(prefix, name))) != null) return sub;
		return null;
	}
	
	/**
	 * Returns the next sibling
	 * 
	 * @return the next sibling
	 */
	public ItsTree<T> nextSibling() {
		return next;
	}
	
	/**
	 * Returns true, if it has a next sibling
	 * @return
	 */
	public boolean hasNextSibling() {
		return next != null;
	}
	
	/**
	 * prints the tree with an intendation level
	 * 
	 * @return the tree as string
	 */
	public String print() {
		StringBuffer s = new StringBuffer();
//		s.append(EOL);
		if(head.isText()) {
			s.append(head.getText());
			return s.toString();
		}
		s.append(head.start(!hasChildren()));
		if(!hasChildren()) return s.toString();
		for(ItsTree<T> child : children) {
			s.append(child.print());
		}
		s.append(head.end());
		return s.toString();
	}
	
	@Override
	public String toString() {
		return getHead().toString();
	}
	
}
