 /* ===========================================================================
 *   Copyright (C) 2013 by ]init[ AG für digitale Kommunikation
 * -----------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  See also the full LGPL text here: http://www.gnu.org/copyleft/lesser.html
 * ===========================================================================*/
package de.init.its.spi;

import java.io.File;
import java.io.Serializable;
import java.net.URL;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;

import de.init.its.api.ds.IitsAttribute;
import de.init.its.api.spi.IitsAttributeManager;
import de.init.its.api.spi.ItsError;
import de.init.its.ds.ItsAttribute;
import de.init.its.ds.ItsAttributes;
import de.init.its.exception.UnsupportedItsAttributeException;
import de.init.its.util.Util;

/**
 * Implementation of {@linkplain IitsAttributeManager}
 * 
 * @author mkandora
 * 
 */
public class ItsAttributeManager implements IitsAttributeManager, Serializable {

	private static final long serialVersionUID = 622513728701971754L;
	
	private static final String DEFAULT_ITS_CONFIG_FILE 	= "/itsconfig.xml";
	private static final String XPATH_ITS_CONFIG 			= "/config/itsconf/*";
	private static final String PREFIX 						= "its:";
	
	private static final transient ItsError NOERROR = new ItsError() {
		
		@Override
		public boolean hasErrors() {
			return false;
		}
		
		@Override
		public String getCause() {
			return "";
		}
	};
	
	private static Map<String, IitsAttributeManager> mgrs = 
			new HashMap<String, IitsAttributeManager>();
	
	private Map<String, IitsAttribute> attributes;

	private ItsAttributeManager(String configFileName) {
		attributes = new HashMap<String, IitsAttribute>();
		File configFile = new File(System.class.getResource(configFileName)
				.getFile());
		try {
			Document conf = DocumentBuilderFactory.newInstance()
					.newDocumentBuilder().parse(configFile);
			XPath xpath = XPathFactory.newInstance().newXPath();
			XPathExpression expr = xpath.compile(XPATH_ITS_CONFIG);
			NodeList nl = (NodeList) expr
					.evaluate(conf, XPathConstants.NODESET);
			for (int i = 0; i < nl.getLength(); i++) {
				NamedNodeMap attr = nl.item(i).getAttributes();
				attributes.put(attr.getNamedItem("name").getNodeValue(),
						new ItsAttribute(attr.getNamedItem("name")
								.getNodeValue(), null));
			}
		} catch (Exception e) {
		} 
	}
	
	private ItsAttributeManager() {
		// load up hardcoded default configuration, if loading of external configuration fails
		attributes = new HashMap<String, IitsAttribute>();
		for(ItsAttributes a : ItsAttributes.values()) {
			attributes.put(a.getAttributeName(), new ItsAttribute(a.getAttributeName(), a.getDefaultValue()));
		}
	}
		

	/**
	 * Returns a singleton instance of the IitsAttributeManager
	 * 
	 * @param configuration file in classpath
	 * @return a singleton instance of the ItsAttributeManager
	 */
	public static IitsAttributeManager getManager(final String configFileName) {
		IitsAttributeManager mgr = null;
		if((mgr = mgrs.get(configFileName)) == null) {
			if(!check(configFileName)) {
				mgr = new ItsAttributeManager();
				mgrs.put(configFileName, mgr);
			}
			else {
				mgr = new ItsAttributeManager(configFileName);
				mgrs.put(configFileName, mgr);
			}
		}
		return mgr;
	}
	
	/**
	 * returns true if config file exists, false otherwise (obviously)
	 * 
	 * @param configFileName the name of configuration file
	 * @return true, if configFile exists
	 */
	private static boolean check(final String configFileName) {
		URL url = System.class.getResource(configFileName);
		if(url == null) return false;
		File configFile = new File(url.getFile());
		return configFile.exists();
	}
	
	/**
	 * Returns a singleton instance of the ItsAttributeManager using the default configuration file
	 * @return
	 */
	private static IitsAttributeManager getManager() {
		return getManager(DEFAULT_ITS_CONFIG_FILE);
	}

	@Override
	public IitsAttribute getAttribute(String itsAttributeName)
			throws UnsupportedItsAttributeException {
		if(itsAttributeName.startsWith(PREFIX)) {
			// strip prefix
			itsAttributeName = itsAttributeName.substring(PREFIX.length());
		}
		
		IitsAttribute attr = attributes.get(itsAttributeName);
		if (attr == null)
			throw new UnsupportedItsAttributeException();
		return attr.copy();
	}
	
	@Override
	public IitsAttribute getAttribute(ItsAttributes attribute)
			throws UnsupportedItsAttributeException {
		return getAttribute(attribute.getAttributeName());
	}

	@Override
	public Collection<IitsAttribute> getAllAvailableAttributes() {
		return attributes.values();
	}
	
//	public static void main(String[] args) {
//		System.out.println(getManager().getAllAvailableAttributes());
//	}

	@Override
	public ItsError validate(final String itsAttributeName, final String itsAttributeValue) {
		ItsAttributes itsAt = null;
		
		for(ItsAttributes i : ItsAttributes.values()) {
			if(i.getAttributeName().equals(itsAttributeName)) {
				itsAt = i;
				break;
			}
		}
		
		if(itsAt == null) {
			return new ItsError() {
				
				@Override
				public boolean hasErrors() {
					return true;
				}
				
				@Override
				public String getCause() {
					return Util.format("provided its attribute [{}] has not been defined and is therefore invalid!", itsAttributeName);
				}
			};
		}
		else if(!itsAt.validate(itsAttributeValue)) {
			return new ItsError() {
				
				@Override
				public boolean hasErrors() {
					return true;
				}
				
				@Override
				public String getCause() {
					return Util.format("provided its argument [{}] for its data category is invalid!", itsAttributeValue, itsAttributeName);
				}
			};
		}
		return NOERROR;
	}
}
