 /* ===========================================================================
 *   Copyright (C) 2013 by ]init[ AG für digitale Kommunikation
 * -----------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  See also the full LGPL text here: http://www.gnu.org/copyleft/lesser.html
 * ===========================================================================*/
package de.init.its.spi;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;
import java.util.concurrent.atomic.AtomicInteger;

import de.init.its.api.ds.IitsAttribute;
import de.init.its.api.ds.IitsId;
import de.init.its.api.ds.IitsTagElement;
import de.init.its.api.process.IitsParserConfig;
import de.init.its.api.spi.IitsAttributeManager;
import de.init.its.api.spi.IitsElementManager;
import de.init.its.api.spi.ItsError;
import de.init.its.ds.ItsAttribute;
import de.init.its.ds.ItsAttributes;
import de.init.its.ds.ItsDocument;
import de.init.its.ds.ItsId;
import de.init.its.ds.ItsTagElement;
import de.init.its.spi.ItsElementManager.ItsTagElementIterator.Executor;
import de.init.its.util.Util;

/**
 * ITS DataCategory management class. Use this class
 * for all ITS DC Modification Operations. The manager
 * keeps a reference to a {@linkplain ItsDocument}, and
 * all available IitsTagElement(s). The internally used structure
 * is tree-like and reflects the nesting of multiple tags inside an
 * odt document file.
 * 
 * 
 * @author mkandora
 * 
 */
public class ItsElementManager implements IitsElementManager {

	private final String ID_PREFIX = "its-";
	
	// MK removed final 
	private static ItsElementManager manager = new ItsElementManager();
	
	private Map<String, IitsTagElement> mapping;
	private ItsDocument itsDocument;
	private IitsTagElement root;
	private ItsTagElementIterator tagIterator;
	private List<ItsError> errors;
	
	private AtomicInteger nextId;
	
	// override OpenOffice mechanism for style persistence
	// and use this field to define an elements attribute
	// to retrieve the id of an ITS tagged element
	private String idReferenceAttributeName;

//	private transient IitsAttributeManager attribMgr;
	
	
	protected ItsElementManager() {
		this.mapping = new HashMap<String, IitsTagElement>();
		this.root		= createRoot();
		this.itsDocument = new ItsDocument(this.root);
		this.tagIterator = new ItsTagElementIterator();
		
		this.nextId 	= new AtomicInteger(0);
		
		// sets the default case for id attributes
		// old configuration
		this.idReferenceAttributeName = IitsParserConfig.ODT_FILE_BOOKMARK_BEGIN_ELEMENT_NAME_NS;
		
		// build error list
		this.errors = new ArrayList<ItsError>();
		
//		this.attribMgr = ItsAttributeManager.getManager();
	}
	
	/**
	 * Obtains a singleton instance of the its element manager
	 * 
	 * @return
	 */
	public static ItsElementManager getManager() {
		return manager;
	}
	
	public static ItsElementManager getManager(final boolean init) {
//		if(init) System.err.println("ItsElementManager initialized...");
		return init ? (manager = new ItsElementManager() ): manager;
	}
	
	public void reset() {
		this.nextId.getAndSet(0);
	}
	
	@Override
	public synchronized void updateId(final String id) throws Exception {
		if(id == null || !id.startsWith(ID_PREFIX) || !id.substring(ID_PREFIX.length()).matches("\\d+")) {
			throw new Exception(Util.format("Illegal ID Format {} should be {} where X is a digit: ", id, "its-XXXXXXX"));
		}
		Integer int_id = Integer.parseInt(id.substring(ID_PREFIX.length()));
		this.nextId.getAndSet(int_id);
	}

	/**
	 * Creates a root {@linkplain IitsTagElement} for internal use
	 * 
	 * @return an {@link IitsTagElement}
	 */
	private IitsTagElement createRoot() {
		IitsTagElement _root = new ItsTagElement();
		_root.setParent(true);
		_root.setParentElement(null);
		return _root;
	}

	@Override
	public IitsTagElement getItsElement(String id) {
		return mapping.get(id);
	}
	
	@Override
	public IitsTagElement insert(ItsAttributes attrib, String... params) {
		return insert("", attrib.getAttributeName(), params[0]);
	}
	
	@Override
	public IitsTagElement insert(final IitsId id, final ItsAttributes dcname, final String dcvalue) {
		return insert(id.getString(), dcname.getAttributeName(), dcvalue);
	}
	
	@Override
	public IitsTagElement insert(final String id, final String dcname, final String dcvalue) {
		String copy = dcvalue;
		IitsTagElement parent = (id != null && !id.isEmpty()) ? findById(id) : this.root;
		
		
//		IitsAttributeManager attribMgr = ItsAttributeManager.getManager();
		
//		ItsError error = attribMgr.validate(dcname,dcvalue);
//		if(error.hasErrors()) {
//			errors.add(error);
//			return null;
//		}
		
		/// check for locale filter list
		if(ItsAttributes.localeFilterList.name().equals(dcname)) {
			copy = copy.trim();
			String[] ar = copy.split(",");
			if(ar.length == 1) {
				copy = ar[0];
			}
		}
		
		
		IitsTagElement nextElement = new ItsTagElement();
		nextElement.addAttribute(new ItsAttribute(dcname, dcvalue));
		String strId = Util.join(ID_PREFIX, getNextId());
		nextElement.setId(new ItsId(strId));
		parent.addChild(nextElement);
		mapping.put(strId, nextElement);
		return nextElement;
	}
	
	@Override
	public IitsTagElement insert(final String dcname, final String dcvalue) {
		return insert("", dcname, dcvalue);
	}
	
	@Override
	public IitsTagElement insert(Map<ItsAttributes, List<String>> values) {
		IitsTagElement e = null;
		for(ItsAttributes its : values.keySet()) {
			if(e == null) {
				e = insert(its, values.get(its).toArray(new String[0]));
				continue;
			}
//			e.addAttribute(its, values.get(its).toString().replaceAll("(\\[|\\])", ""));
			e.addAttribute(its, values.get(its).get(0));
		}
		return e;
	}
	
	@Override
	public IitsTagElement remove(final String id) {
		IitsTagElement element = mapping.remove(id);
		
		Stack<IitsTagElement> stack = new Stack<IitsTagElement>();
		stack.push(root);
		IitsTagElement next;
		IitsTagElement parent;
		IitsTagElement child;
		try {
			while(!stack.isEmpty())
			{
				next = stack.pop();
				if(next.getId().getString().equals(id)) {
					// remove this element from the tree ...
					parent = next.getParentElement();
					parent.removeChild(next);
					while(next.hasMoreChildren()) {
						child = next.getNextChild();
						parent.addChild(child);
					}
					// .. and skip the iteration
					break;
				} else {
					if(next.hasChildren()) {
						while(next.hasMoreChildren()) {
							stack.push(next.getNextChild());
						}
					}
				}
			}
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return element;
	}
	
	@Override
	public List<IitsTagElement> findAll() {
		List<IitsTagElement> list = new ArrayList<IitsTagElement>();

		Executor<IitsTagElement> exec = new Executor<IitsTagElement>() {
			@Override
			public void execute(IitsTagElement t, List<IitsTagElement> container) {
				container.add(t);
			}
		};
		tagIterator.iterate(itsDocument, exec, list);
		return list;
	}

	@Override
	public List<IitsTagElement> findByAttribute(final IitsAttribute attribute) {
		List<IitsTagElement> list = new ArrayList<IitsTagElement>();
		Executor<IitsTagElement> exec = new Executor<IitsTagElement>() {
			@Override
			public void execute(IitsTagElement t, List<IitsTagElement> container) {
				if(t.getAttributes().contains(attribute)) container.add(t);
			}
		};
		tagIterator.iterate(itsDocument, exec, list);
		return list;
	}

	@Override
	public List<IitsTagElement> findByAttributeName(final String attributeName) {
		List<IitsTagElement> list = new ArrayList<IitsTagElement>();
		Executor<IitsTagElement> exec = new Executor<IitsTagElement>() {
			@Override
			public void execute(IitsTagElement t, List<IitsTagElement> container) {
				for(IitsAttribute a : t.getAttributes()) {
					if(attributeName.equals(a.getItsAttributeName()))
						container.add(t);
				}
			}
		};
		tagIterator.iterate(itsDocument, exec, list);
		return list;
	}
	
	@Override
	public IitsTagElement findById(final String id) {
		
		List<IitsTagElement> list = new ArrayList<IitsTagElement>();
		Executor<IitsTagElement> exec = new Executor<IitsTagElement>() {
			@Override
			public void execute(IitsTagElement t, List<IitsTagElement> container) {
					if(id.equals(t.getId().getString())) {
						container.add(t);
					}
			}
		};
		tagIterator.iterate(itsDocument, exec, list);
		return !list.isEmpty() ? list.get(0) : null; 
	}
	
	@Override
	public IitsTagElement findById(final IitsId id) {
		return findById(id.getString());
	}
	
	@Override
	public String getIdReferenceAttributeName() {
		return idReferenceAttributeName;
	}
	
	@Override
	public void setIdReferenceAttributeName(final String name) {
		this.idReferenceAttributeName = name;
	}
	
	/**
	 * returns the next available id
	 * 
	 * @return the next id as Integer
	 */
	protected synchronized Integer getNextId() {
		return nextId.incrementAndGet();
	}
	
	@Override
	public void setItsDocument(ItsDocument document) {
		this.itsDocument = document;
		this.root = this.itsDocument.getRoot();
	}
	/**
	 * A generic ITS tag element iterator
	 * 
	 * @author mkandora
	 *
	 */
	protected static class ItsTagElementIterator implements Serializable {
		protected static interface Executor<T> {
			public void execute(final T t, final List<T> container);
		}
		
		private boolean terminate;
	
		/**
		 * Iterates over an {@link IitsTagElement} tree using an executor 
		 * 
		 * @param doc the {@link ItsDocument} containing the {@link IitsTagElement} root tree
 		 * @param exec an {@link Executor}
		 * @param list a list for values
		 */
		public void iterate(final ItsDocument doc,
				final Executor<IitsTagElement> exec,
				final List<IitsTagElement> list) {
			IitsTagElement next = null;
			Stack<IitsTagElement> stack = new Stack<IitsTagElement>();
			stack.push(doc.getRoot());
			while (!stack.isEmpty() /*&& !isTerminated() */) {
				next = stack.pop();
				exec.execute(next, list);
				while (next.hasMoreChildren()) {
					stack.push(next.getNextChild());
				}
				next.reset();
			}
		}
		
		/**
		 * returns true, if iteration should be terminated
		 * @return
		 */
		protected boolean isTerminated() {
			return terminate;
		}
		
		/**
		 * set this flag to terminate iteration
		 * @param terminate
		 */
		protected void setTerminate(boolean terminate) {
			this.terminate = terminate;
		}
		
	}
	@Override
	public ItsDocument getItsDocument() {
		return itsDocument;
	}

	@Override
	public List<ItsError> getErrors() {
		return errors;
	}

	@Override
	public boolean hasErrors() {
		return !errors.isEmpty();
	}
	
	@Override
	public void clearErrors() {
		errors.clear();
	}
}
