 /* ===========================================================================
 *   Copyright (C) 2013 by ]init[ AG für digitale Kommunikation
 * -----------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  See also the full LGPL text here: http://www.gnu.org/copyleft/lesser.html
 * ===========================================================================*/
package de.init.its.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.StringReader;
import java.net.URI;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.charset.Charset;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.xml.sax.InputSource;

/**
 * Support class to implement overall need functionality 
 * 
 * @author mkandora
 *
 */
public final class Util {

	/**
	 * The HOME directory java is working in
	 */
	public static final String HOME = System.getProperty("user.dir");
	
	private Util(){}

	/**
	 * Returns a textual file as InputSource
	 * 
	 * @param resource the absolute path to a resource
	 * @return {@linkplain InputSource}
	 * @throws Exception 
	 */
	public static InputSource toInputSource(final String resource) throws Exception {
		File f = new File(resource);
		if(!f.exists()) throw new FileNotFoundException();
		FileInputStream fis = new FileInputStream(f);
		FileChannel channel = fis.getChannel();
		MappedByteBuffer map = channel.map(FileChannel.MapMode.READ_ONLY, 0, channel.size());
		fis.close();
		return new InputSource(new StringReader(Charset.defaultCharset().decode(map).toString()));
	}
	
	/**
	 * Copies a file
	 * 
	 * @param uri
	 * @param suffix
	 * @return
	 */
	public static URI mkTmpCopy(final URI uri, final String suffix) {
	    FileChannel src = null;
	    FileChannel dst = null;
	    File tmpFile = null;
	    try {
	    	tmpFile = File.createTempFile("copy", suffix);
	        src = new FileInputStream(uri.getPath()).getChannel();
	        
			dst = new FileOutputStream(tmpFile).getChannel();
	        dst.transferFrom(src, 0, src.size());
	        if(src != null) {
	            src.close();
	        }
	        if(dst != null) {
	            dst.close();
	        }
	    } 
	    catch(Exception e) {
	    }
	    return tmpFile.toURI();
	}
	
	/**
	 *  Returns a textual file as InputSource
	 *  
	 * @param resource
	 * @return
	 * @throws Exception
	 */
	public static InputSource toInputSource(final File resource) throws Exception {
		return toInputSource(resource.getPath());
	}
	
	/**
	 * Reads a zipFile and returns a specific Zip Entry as InputSource and returns it.
	 * @param resource  the zip file
	 * @param zipEntry the entry to be returned as InputSource
	 * @return
	 * @throws Exception
	 */
	public static InputSource toInputSourceFromZip(final String resource, final String zipEntry) throws Exception {
		ZipFile zip = new ZipFile(resource);
		return new InputSource(
				zip.getInputStream(
						zip.getEntry(zipEntry)));
	}
	
	/**
	 * Extracts an entry from a zip file and saves it as temporary resource.
	 * returns a file reference to the temp resource
	 * 
	 * @param zipResource the zip file (like an odt file)
	 * @param entryName the name of the entry to extract
	 * @return the temporary resource
	 * @throws Exception
	 */
	public static File extractZipEntry(final String zipResource, final String entryName) throws Exception {
		String path = zipResource.substring(0, zipResource.lastIndexOf("/") + 1);
		ZipFile zfile = new ZipFile(zipResource);
		ZipEntry entry = zfile.getEntry(entryName);
		InputStream is = zfile.getInputStream(entry);
		File tmp = File.createTempFile("context", "xml");
		FileOutputStream tfos  = new FileOutputStream(tmp);
		byte[] buffer = new byte[0x1000];
		int len = -1;
		while((len = is.read(buffer)) != -1) {
			tfos.write(buffer, 0, len);
		}
		tfos.flush();
		tfos.close();
		is.close();
		return tmp;
	}
	
	
	
	/**
	 * Convenience Method to convert a string represenation of an xml document
	 * to an {@linkplain InputSource} 
	 * @param plainXMLRepresenation
	 * @return
	 * @throws Exception
	 */
	public static InputSource convertFrom(final String plainXMLRepresenation) throws Exception {
		return new InputSource(new StringReader(plainXMLRepresenation));
	}
	
	/**
	 * Joins a String with given Object parameters 
	 * 
	 * @param a a String as start
	 * @param b various elements to join with String a
	 * @return
	 */
	public static String join(String a, Object...b) {
		StringBuffer bu = new StringBuffer(a);
		for (int i = 0; i < b.length; i++) bu.append(b[i]);
		return bu.toString();
	}
	
	/**
	 * Returns true if the map contains a key with the provided prefix
	 * 
	 * @param map the map to check for a key with a given prefix
	 * @param prefix the prefix to check for existence
	 * @return true, if key with prefix exists
	 */
	public static boolean containsKeyPrefix(final Map<String,String> map, final String prefix) {
		for(final String key : map.keySet()) {
			if(key.startsWith(prefix)) return true;
		} return false;
	}
	
	/**
	 * crops a string according to start and end (both exclusive) delimiters.
	 * if input string does not match both delimiters the input string will be returned
	 * unmodified.
	 * 
	 * @param input the input to be cropped
	 * @param begin the start delimiter
	 * @param end the end delimiter
	 * @return the cropped string
	 */
	public static String crop(final String input, String begin, String end) {
		if(!input.matches(".*\\"+ begin + ".*\\"+ end +"+")) return input;
		return input.substring(input.indexOf(begin) + 1, input.lastIndexOf(end));
	}
	
	
	/**
	 * tokenizes a string by delimiter and returns the n th field (begin index with 0)
	 * 
	 * @param str the string to be cut
	 * @param delim the delimiter to split the string
	 * @param field the field, that needs to be returned
	 * @return the nth field of the string
	 */
	public static String cut(final String str, final String delim, final int field) {
		String[] fields = str.split(escape(delim));
		if(field >= fields.length || field < 0) return "";
		return fields[field];
	}
	
	/**
	 * Escapes almost all possible occurences of:
	 * 
	 * <ul>
	 * <li>\t</li>
	 * <li>\r</li>
	 * <li>\n</li>
	 * <li>\s</li>
	 * <li>\d</li>
	 * <li>:</li>
	 * <li>,</li>
	 * <li>-</li>
	 * <li>=</li>
	 * </ul>
	 *
	 * @param str the string that may contain possible control characters
	 * @return the fully escaped string
	 */
	public static String escape(final String str) {
		StringBuilder sb = new StringBuilder();
		for(char c : str.toCharArray()) {
			switch(c) {
			case '\t':
				sb.append("\\t");
				break;
			case '\n':
				sb.append("\\n");
				break;
			case '\r':
				sb.append("\\r");
				break;
			case ':':
				sb.append("\\:");
				break;
			case '=':
				sb.append("\\=");
				break;
			case '-':
				sb.append("\\-");
				break;
			case '.':
				sb.append("\\.");
				break;
			case '|':
				sb.append("\\|");
				break;
			default:
				sb.append(c);
			}
		}
		return sb.toString();
	}
	
	/**
	 * Formats a given String. Backward Compatibility to Support Java 1.5 Projects
	 * See {@linkplain String#format(String, Object...)} for Java 1.6
	 * 
	 * @param s the simple string with {} as replacement symbol
	 * @param o a variable number of parameters
	 * @return the formatted String
	 */
	public static String format(final String s, Object...o) {
		String variable = "\\{\\}";
		StringBuffer sb = new StringBuffer();
		int index = 0;
		Matcher m = Pattern.compile(variable).matcher(s);
		while(m.find()) {
			Object rplc = ""; 
			if(index < o.length) {
				rplc = o[index];
			}
			m.appendReplacement(sb, "" + rplc);
			index++;
		}
		m.appendTail(sb);
		return sb.toString();
	}
	
	/**
	 * Returns true, if any of the supplied arguments are contained in the map
	 * @param map the map to look for key entries
	 * @param args the arguments to match any of the key entries
	 * @return true, if a single argument matches an entry in the map
	 */
	public static <T> boolean matchesAny(final Map<String, String> map, String...args) {
		if(map == null) return false;
		for(String a : args) {
			if(map.containsKey(a)) return true;
		} return false;
	}
	
	/**
	 * Uses a boolean delegate to evaluate all arguments, returns true
	 * if one of the arguments matches a condition defined by the delegate
	 *  
	 * @param executor a delegate to match a single argument in list
	 * @param list a list of argument values
	 * @return true, if one argument matches the condition by the delegate
	 */
	public static <T> boolean matchSingle(MatcherOp<T> executor, List<T> list) {
		for(T t : list) { 
			if(executor.match(t)) return true;
		}
		return false;
	}
	
	/**
	 * Extracts the Path to a File (eg. the Parent Dir)
	 * 
	 * @param path
	 * @return
	 */
	public static String extractPath(final String path) {
		return path.substring(0, path.lastIndexOf('/'));
	}
	
	/**
	 * Writes string content to disk
	 * @param path
	 * @param content
	 * @throws Exception
	 */
	public static void writeToDisk(final String path, final String content) throws Exception {
		File file = new File(path);
		if(!file.exists()) file.createNewFile();
		BufferedWriter writer = new BufferedWriter(new FileWriter(file));
		writer.write(content);
		writer.flush();
		writer.close();
	}
	
	/**
	 * Repeats a string n times and returns it
	 * @param sym
	 * @param times
	 * @return
	 */
	public static String rep(final String sym, final int times) {
		return new String(new char[times]).replaceAll("\0", sym);
	}
	/**
	 * returns a file reference for a resource
	 * 
	 * @param resource
	 * @return
	 * @throws FileNotFoundException
	 */
	public static File toFile(final String resource) throws FileNotFoundException {
		File target = new File(resource);
		if(!target.exists()) throw new FileNotFoundException("File does not exist");
		return target;
	}
	
	/**
	 * Returns the String representation of a File. Might only be useful for
	 * textual files.
	 * 
	 * @param file the source file
	 * @return the String content of the file
	 * @throws Exception
	 */
	public static String getStringContent(final File file) throws Exception {
		StringBuffer buffer = new StringBuffer();
		BufferedReader reader = new BufferedReader(new FileReader(file));
		while(reader.ready()) {
			buffer.append(reader.readLine());
		}
		reader.close();
		return buffer.toString();
	}
	
	public static URI prepreXLIFFOutput() {
		
		return null;
	}
	
	/**
	 * Calls System.out.printline with a more advanced formatting of objects 
	 * (Backward compatibility for Java 5)
	 * 
	 * @param string the string containing {} as placeholder for values
	 * @param format the formatting values, can be empty (not null)
	 */
	public static void println(final String string, final Object...format) {
		System.out.println(Util.format(string, format));
	}
	
	/**
	 * Simple Boolean Matcher Executor Interface
	 * 
	 * @author mkandora
	 * @param <T>
	 */
	public static interface MatcherOp<T> {
		public boolean match(final T input);
	}
	
	/**
	 * Simple executor delegator interface
	 * 
	 * @author mkandora
	 *
	 * @param <T>
	 */
	public static interface Executor<T> {
		public void execute(final T input);
	}
	
}
