 /* ===========================================================================
 *   Copyright (C) 2013 by ]init[ AG für digitale Kommunikation
 * -----------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  See also the full LGPL text here: http://www.gnu.org/copyleft/lesser.html
 * ===========================================================================*/
package de.init.its.util;

import java.util.HashMap;
import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

import de.init.its.api.ds.IitsAttribute;
import de.init.its.api.ds.IitsTagElement;
import de.init.its.api.spi.IitsElementManager;
import de.init.its.spi.ItsElementManager;

import static de.init.its.api.process.IitsParserConfig.*;
/**
 * Some useful XML utils
 * 
 * @author mkandora
 *
 */
public class XMLUtil {
	
	/**
	 * Retrieves all namespaces as defined by the root element from given Document
	 * 
	 * Beta function to get all available namespacess from root element.
	 * FIX if there is a better option. 
	 * 
	 * @param doc the document
	 * @param map a map container to store all namespaces
	 */
	public static void fetchNS(final Document doc, final Map<String, String> map) {
		Map<String, String> attributes = extractAttributes(doc.getDocumentElement());
		for(String name : attributes.keySet()) {
			if(name.startsWith(XML_NAMESPACE_PREFIX)) {
				map.put(name, attributes.get(name));
			}
		}
	}
	
	/**
	 * Extracts all attributes from a given node and returns it as a map
	 * @param node the node to retrieve all attributes
	 * @return
	 */
	public static Map<String, String> extractAttributes(Node node) {
		Map<String, String> attributes = new HashMap<String, String>();
		int index = -1;
		NamedNodeMap map = node.getAttributes();
		if(map == null) return attributes;
		while((++index) < map.getLength())
		{
			Node attribute = map.item(index);
			attributes.put(attribute.getNodeName(), attribute.getNodeValue());
		}
		return attributes;
	}
	
	/**
	 * Sets the same attribute for multiple elements
	 * 
	 * @param name the name of the attribute
	 * @param value the value of the attribute
	 * @param node the nodes to set the same attribute for
	 */
	public static void setAttribute(final String name, final String value, final Node...node) {
		for (final Node element : node) {
			((Element)element).setAttribute(name, value);
		}
	}
	
	/**
	 * Sets a namespace attribute 
	 * 
	 * @param namespaceURI
	 * @param name
	 * @param value
	 * @param node
	 */
	public static void setAttributeNS(final String namespaceURI, final String name, final String value, final Node...node) {
		for (final Node element : node) {
			((Element)element).setAttributeNS(namespaceURI, name, value);
		}
	}
	
	/**
	 * Returns true, if element has desired attribute 
	 * 
	 * @param e the element to look for specific attribute
	 * @param attributeName the attribute name 
	 * @return true, if attribute exists
	 */
	public static boolean hasAttribute(final Element e, final String attributeName) {
		int index = -1;
		NamedNodeMap attributes = e.getAttributes();
		while((++index) < attributes.getLength()) {
			if(attributeName.equals(attributes.item(index).getNodeName())) return true;
		} 
		return false;
	}
	
	/**
	 * Returns true, if node has an attribute with provided prefix
	 * @param node the node to look for desired attribute prefix
	 * @param prefix the attribute prefix
	 * @return true, if attribute with prefix exists
	 */
	public static boolean containsAttributeWithPrefix(final Node node, final String prefix) {
		for(String key : extractAttributes(node).keySet()) {
			if(key.startsWith(prefix)) return true;
		}
		return false;
	}
	
	/**
	 * Extracts the its id from a given bookmark node
	 * @param node
	 * @return
	 */
	public static String extractId(final Node node) {
		return Util.crop(extractAttributes(node).get(ODT_FILE_BOOKMARK_REFERENCING_ATTRIBUTE_NAME), "[", "]");
	}
	
	/**
	 * Creates a copy from the given node
	 * @param doc
	 * @param node
	 * @return
	 */
	public static Node copyFrom(final Document doc, final Node node, final String namespaceprefix) {
		Element element = null;
		String pre = namespaceprefix != null ? (namespaceprefix.startsWith("xmlns:") ? namespaceprefix : "xmlns:" + namespaceprefix) : "";
		switch(node.getNodeType()) {
		case Node.ELEMENT_NODE:
			element = doc.createElement(pre + node.getNodeName());
			Map<String, String> map = extractAttributes(node);
			for(String key : map.keySet()) {
				element.setAttribute(key, map.get(key));
			}
			break;
		case Node.TEXT_NODE: return node;
		}
		return element;
	}
	
	/**
	 * Creates a new itsNode
	 * @param doc
	 * @param bookmarkNode
	 * @param manager
	 * @return
	 */
	public static Element createItsNode(final Document doc, final Node bookmarkNode, final ItsElementManager manager) {
		String id = extractId(bookmarkNode);
		Element e = doc.createElement(ODT_FILE_TARGET_ITS_ELEMENT_WITH_PREFIX);
		IitsTagElement its = manager.findById(id);
		for (IitsAttribute itsAttribute : its.getAttributes()) {
			XMLUtil.setAttribute(
					itsAttribute.getItsAttributeName(),
					itsAttribute.getItsAttributeValue(), e);
		}
		return e;
	}
}
